from lcapy import *
import numpy as np
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy state space analysis
    """

    def assertEqual2(self, ans1, ans2, comment):

        ans1 = ans1.canonical()
        ans2 = ans2.canonical()

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            ans1.pprint()
            ans2.pprint()
            raise AssertionError(e)

    def test_VRC1(self):
        """Check VRC circuit

        """
        a = Circuit("""
        V1 1 0
        R1 1 2
        C1 2 0""")

        ss = a.ss

        self.assertEqual2(expr(ss.x[0]), expr(
            'v_C1(t)'), "Incorrect state variable")
        self.assertEqual2(expr(ss.y[0]), expr(
            'v_1(t)'), "Incorrect output variable1")
        self.assertEqual2(expr(ss.y[1]), expr(
            'v_2(t)'), "Incorrect output variable2")
        self.assertEqual2(expr(ss.A[0]), expr(
            '-1/(R1 * C1)'), "Incorrect A matrix")
        self.assertEqual2(expr(ss.B[0]), expr(
            '1/(R1 * C1)'), "Incorrect B matrix")
        self.assertEqual2(expr(ss.C[0]), expr(
            0), "Incorrect C[0] matrix element")
        self.assertEqual2(expr(ss.C[1]), expr(
            1), "Incorrect C[1] matrix element")
        self.assertEqual2(expr(ss.D[0]), expr(
            1), "Incorrect D[0] matrix element")
        self.assertEqual2(expr(ss.D[1]), expr(
            0), "Incorrect D[1] matrix element")
        self.assertEqual2(expr(ss.eigenvalues[0]), expr(
            '-1/(R1 * C1)'), "Incorrect eigenvalues")

    def test_VRL1(self):
        """Check VRL circuit

        """
        a = Circuit("""
        V1 1 0
        R1 1 2
        L1 2 0""")

        ss = a.ss

        self.assertEqual2(expr(ss.x[0]), expr(
            'i_L1(t)'), "Incorrect state variable")
        self.assertEqual2(expr(ss.y[0]), expr(
            'v_1(t)'), "Incorrect output variable1")
        self.assertEqual2(expr(ss.y[1]), expr(
            'v_2(t)'), "Incorrect output variable2")
        self.assertEqual2(expr(ss.A[0]), expr(
            '-R1 / L1'), "Incorrect A matrix")
        self.assertEqual2(expr(ss.B[0]), expr('1 / L1'), "Incorrect B matrix")
        self.assertEqual2(expr(ss.C[0]), expr(
            0), "Incorrect C[0] matrix element")
        self.assertEqual2(expr(ss.C[1]), expr(
            '-R1'), "Incorrect C[1] matrix element")
        self.assertEqual2(expr(ss.D[0]), expr(
            1), "Incorrect D[0] matrix element")
        self.assertEqual2(expr(ss.D[1]), expr(
            1), "Incorrect D[1] matrix element")
        self.assertEqual2(expr(ss.eigenvalues[0]), expr(
            '-R1 / L1'), "Incorrect eigenvalues")

    def test_RLC(self):

        a = Circuit("""
        V 1 0 {v(t)}; down
        R1 1 2; right
        L 2 3; right=1.5, i={i_L}
        R2 3 0_3; down=1.5, i={i_{R2}}, v={v_{R2}}
        W 0 0_3; right
        W 3 3_a; right
        C 3_a 0_4; down, i={i_C}, v={v_C}
        W 0_3 0_4; right""")
        ss = a.ss

        self.assertEqual(ss.x[0], expr('i_L(t)'), "x[0]")
        self.assertEqual(ss.x[1], expr('v_C(t)'), "x[1]")

        self.assertEqual(ss.x0[0], 0, "x0[0]")
        self.assertEqual(ss.x0[1], 0, "x0[1]")

        self.assertEqual(ss.y[0], expr('v_1(t)'), "y[0]")

        self.assertEqual(ss.u[0], expr('v(t)'), "u[0]")

        self.assertEqual(ss.A[0, 0], expr('-R1/L'), "A[0, 0]")

    def test_transfer(self):
        Z = (s**2 + 3) / (s**3 + 2 * s + 10)

        ss = Z.state_space()

        self.assertEqual(ss.Nx, 3, "Nx")
        self.assertEqual(ss.Ny, 1, "Ny")
        self.assertEqual(ss.Nu, 1, "Nu")
        self.assertEqual(ss.is_stable, False, "is_stable")
        self.assertEqual(ss.is_controllable, True, "is_controllable")
        self.assertEqual(ss.is_observable, True, "is_observable")
        self.assertEqual(ss.is_symbolic, False, "is_symbolic")
        self.assertEqual(ss.G[0], Z, "G")

        sso = Z.state_space('OCF')
        self.assertEqual(sso.G[0], Z, "G")

        ssd = Z.state_space('DCF')
        #self.assertEqual(ssd.G[0], Z, "G")

    def test_balanced(self):

        A = [[1, -2], [3, -4]]
        B = [5, 7]
        C = [[6, 8]]
        D = [9]

        ss = StateSpace(A, B, C, D)

        ssb = ss.balance()

        h = ss.hankel_singular_values

        H = np.diag(h.numpy.squeeze())

        self.assertEqual(np.allclose(
            ssb.Wo.numpy, ssb.Wc.numpy), True, "Wo==Wc")
        self.assertEqual(np.allclose(ssb.Wo.numpy, H),
                         True, "Hankel singular values")
        self.assertEqual(ss.eigenvalues, [-1, -2], "eigen values")



import lcapy.grammar as grammar
import lcapy.schemcpts as schemcpts
from lcapy.parser import Parser
from nose.tools import *
import sys
sys.path.append('..')


parser = Parser(schemcpts, grammar)
parse = parser.parse


@raises(ValueError)
def test_Exception1():
    '''Test missing arg'''

    parse('V1 2')


@raises(ValueError)
def test_Exception2():
    '''Test too many args'''

    parse('V1 2 3 4 5 6')


@raises(ValueError)
def test_Exception3():
    '''Test too many args'''

    parse('V1 2 3 dc 4 5 6')


@raises(ValueError)
def test_Exception4():
    '''Test unknown component'''

    parse('B1 1 2')


@raises(ValueError)
def test_Exception5():
    '''Test arg reassignment'''

    parse('E1 1 2 opamp 3 4 A Ad=10')


@raises(ValueError)
def test_Exception6():
    '''Test unknown arg'''

    parse('V1 1 2 foo=3')


def test_V1():
    '''Test voltage source'''

    assert_equals(type(parse('V1 1 2')), schemcpts.V, 'Class not V')


def test_Vdc1():
    '''Test dc voltage source'''

    assert_equals(type(parse('V1 1 2 dc')),
                  schemcpts.classes['Vdc'], 'Class not Vdc')


def test_Vac1():
    '''Test ac voltage source'''

    assert_equals(type(parse('V1 1 2 ac')),
                  schemcpts.classes['Vac'], 'Class not Vac')


def test_Vsin1():
    '''Test sin voltage source'''

    assert_equals(type(parse('V1 1 2 sin(1, 2, 3)')),
                  schemcpts.classes['Vsin'], 'Class not Vsin')


def test_Vexpr():
    '''Test voltage source with expr'''

    assert_equals(type(parse('V1 1 2 {a * 5}')),
                  schemcpts.V, 'Class not V')


def test_Vac2():
    '''Test ac voltage source'''

    assert_equals(
        type(parse('V1 1 2 ac {1j} {0} 3')), schemcpts.classes['Vac'], 'Class not Vac')


def test_Vquotes():
    '''Test voltage source with arg in quotes'''

    assert_equals(type(parse('V1 1 2 "a * 5"')),
                  schemcpts.V, 'Class not V')

# def test_opamp():
#     '''Test opamp'''
#
#     assert_equals(type(parse('E 1 2 opamp 3 4')), schemcpts.classes['Eopamp'], 'Class not Eopamp')



from lcapy import *
from lcapy.units import as_value_unit, u as uu
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_time_units(self):

        self.assertEqual(str(voltage(texpr(1)).units), 'V', 'time voltage')
        self.assertEqual(str(current(texpr(1)).units), 'A', 'time current')
        self.assertEqual(str(admittance(texpr(1)).units),
                         'S/s', 'time admittance')
        self.assertEqual(str(impedance(texpr(1)).units),
                         'ohm/s', 'time impedance')
        self.assertEqual(str((voltage(texpr(1))**2).units),
                         'V**2', 'time voltage squared')
        self.assertEqual(str((current(texpr(1))**2).units),
                         'A**2', 'time current squared')
        self.assertEqual(str((admittance(texpr(1))**2).units),
                         'S**2/s**2', 'time admittance squared')
        self.assertEqual(str((impedance(texpr(1))**2).units),
                         'ohm**2/s**2', 'time impedance squared')

    def test_fourier_units(self):

        self.assertEqual(str(voltage(fexpr(1)).units),
                         'V/Hz', 'fourier voltage')
        self.assertEqual(str(current(fexpr(1)).units),
                         'A/Hz', 'fourier current')
        self.assertEqual(str(admittance(fexpr(1)).units),
                         'S', 'fourier admittance')
        self.assertEqual(str(impedance(fexpr(1)).units),
                         'ohm', 'fourier impedance')
        self.assertEqual(str((voltage(fexpr(1))**2).units),
                         'V**2/Hz**2', 'fourier voltage squared')
        self.assertEqual(str((current(fexpr(1))**2).units),
                         'A**2/Hz**2', 'fourier current squared')
        self.assertEqual(str((admittance(fexpr(1))**2).units),
                         'S**2', 'fourier admittance squared')
        self.assertEqual(str((impedance(fexpr(1))**2).units),
                         'ohm**2', 'fourier impedance squared')

    def test_constant_units(self):

        self.assertEqual(str(voltage(cexpr(1)).units), 'V', 'constant voltage')
        self.assertEqual(str(current(cexpr(1)).units), 'A', 'constant current')
        self.assertEqual(str(admittance(cexpr(1)).units),
                         'S', 'constant admittance')
        self.assertEqual(str(impedance(cexpr(1)).units),
                         'ohm', 'constant impedance')
        self.assertEqual(str((voltage(cexpr(1))**2).units),
                         'V**2', 'constant voltage squared')
        self.assertEqual(str((current(cexpr(1))**2).units),
                         'A**2', 'constant current squared')
        self.assertEqual(str((admittance(cexpr(1))**2).units),
                         'S**2', 'constant admittance squared')
        self.assertEqual(str((impedance(cexpr(1))**2).units),
                         'ohm**2', 'constant impedance squared')

        self.assertEqual(str(resistance(1).units), 'ohm', 'resistance')
        self.assertEqual(str(conductance(1).units), 'S', 'conductance')
        self.assertEqual(str(capacitance(1).units), 'F', 'capacitance')
        self.assertEqual(str(inductance(1).units), 'H', 'inductance')
        self.assertEqual(str(reactance(1).units), 'ohm', 'reactance')
        self.assertEqual(str(susceptance(1).units), 'S', 'susceptance')

    def test_phasor_units(self):

        self.assertEqual(str(voltage(phasor(1)).units), 'V', 'phasor voltage')
        self.assertEqual(str(current(phasor(1)).units), 'A', 'phasor current')
        self.assertEqual(str(admittance(phasor(1)).units),
                         'S', 'phasor admittance')
        self.assertEqual(str(impedance(phasor(1)).units),
                         'ohm', 'phasor impedance')
        self.assertEqual(str((voltage(phasor(1))**2).units),
                         'V**2', 'phasor voltage squared')
        self.assertEqual(str((current(phasor(1))**2).units),
                         'A**2', 'phasor current squared')
        self.assertEqual(str((admittance(phasor(1))**2).units),
                         'S**2', 'phasor admittance squared')
        self.assertEqual(str((impedance(phasor(1))**2).units),
                         'ohm**2', 'phasor impedance squared')

    def test_discrete_time_units(self):

        self.assertEqual(str(voltage(nexpr(1)).units), 'V', 'nexpr voltage')
        self.assertEqual(str(current(nexpr(1)).units), 'A', 'nexpr current')
        self.assertEqual(str(admittance(nexpr(1)).units),
                         'S', 'nexpr admittance')
        self.assertEqual(str(impedance(nexpr(1)).units),
                         'ohm', 'nexpr impedance')
        self.assertEqual(str((voltage(nexpr(1))**2).units),
                         'V**2', 'nexpr voltage squared')
        self.assertEqual(str((current(nexpr(1))**2).units),
                         'A**2', 'nexpr current squared')
        self.assertEqual(str((admittance(nexpr(1))**2).units),
                         'S**2', 'nexpr admittance squared')
        self.assertEqual(str((impedance(nexpr(1))**2).units),
                         'ohm**2', 'nexpr impedance squared')

    def test_discrete_fourier_units(self):

        self.assertEqual(str(voltage(kexpr(1)).units), 'V', 'kexpr voltage')
        self.assertEqual(str(current(kexpr(1)).units), 'A', 'kexpr current')
        self.assertEqual(str(admittance(kexpr(1)).units),
                         'S', 'kexpr admittance')
        self.assertEqual(str(impedance(kexpr(1)).units),
                         'ohm', 'kexpr impedance')
        self.assertEqual(str((voltage(kexpr(1))**2).units),
                         'V**2', 'kexpr voltage squared')
        self.assertEqual(str((current(kexpr(1))**2).units),
                         'A**2', 'kexpr current squared')
        self.assertEqual(str((admittance(kexpr(1))**2).units),
                         'S**2', 'kexpr admittance squared')
        self.assertEqual(str((impedance(kexpr(1))**2).units),
                         'ohm**2', 'kexpr impedance squared')

    def test_new_units(self):

        v = voltage(3)
        i = current(4)

        p = v * i
        # Convert V * W to W if canonical_units not enabled.
        p = p.simplify_units()

        self.assertEqual(str(p.units), 'W', 'power')

    def test_functions(self):

        self.assertEqual(str(delta(t).units), '1/s', 'delta(t)')
        self.assertEqual(str(diff(t, t).units), '1', 'diff(t, t)')
        self.assertEqual(str(integrate(t, t).units), 's**2', 'integrate(t, t)')
        self.assertEqual(str(integrate(t, (t, 0, 1)).units),
                         's**2', 'integrate(t, (t, 0, 1))')
        self.assertEqual(str(atan(f).units), 'rad', 'atan(f)')

    def test_convolution(self):

        i = current('i(t)')
        z = impedance('z(t)')
        v = i.convolve(z).simplify_units()

        self.assertEqual(str(v.units), 'V', 'i convolve z')

    def test_as_value_unit(self):

        i = current(7)
        val, unit = as_value_unit(i.expr_with_units)
        self.assertEqual(str(unit), 'A', 'test unit')
        self.assertEqual(val, 7, 'test val')



"""This module provides the Latex class for formatting
values and labels.

Copyright 2020--2022 Michael Hayes, UCECE

"""

import re
from .config import subscripts, greek_letter_names

sub_super_pattern = re.compile(r"([_\^]){([a-zA-Z]+)([0-9]*)}")
greek_letter_name_pattern = re.compile(
    r"([^a-zA-Z\\]|^)(%s)([^a-zA-Z]|$)" % '|'.join(greek_letter_names))
math_symbols_pattern = re.compile(
    '|'.join(['_',  r'\^', r'\\left', r'\\math', r'\\frac', r'\\sqrt']))


class Latex(object):

    def __init__(self, string):

        self.str = string

    def mathrm(self, s):
        """Place words in sub- or super-scripts inside a mathrm.
        For example: V_{rms} -> V_{\mathrm{rms}}"""

        def foo(match):

            word = match.group(2)
            suffix = word + match.group(3)

            # Perhaps look up dictionary to find valid words?
            # Assume that if length 3 or more then a word.
            if word.lower() in subscripts or len(word) > 2:
                suffix = r'{\mathrm{%s}}' % suffix
            else:
                suffix = r'{%s}' % suffix
            return match.group(1) + suffix

        return sub_super_pattern.sub(foo, s)

    def greek(self, s):
        """Replace alpha with \alpha, etc."""

        def foo(match):
            return match.group(1) + '\\' + match.group(2) + match.group(3)

        return greek_letter_name_pattern.sub(foo, s)

    def __str__(self):

        s = self.str
        s = self.greek(s)
        return self.mathrm(s)


def latex_str(string):

    return Latex(string).__str__()


def latex_format_label(s):

    if s == '':
        return s

    # With leading $ and no trailing $, e.g., v=$V1, try to
    # automagically convert to LaTeX string, otherwise pass
    # verbatim.  Currently, this only converts words in sub- or
    # super- scripts to roman.  TODO, handle more cases.
    if s[0] == '$' and s[-1] != '$':
        return '$' + latex_str(s[1:]) + '$'

    # If find a $ assume that the label is correctly formatted.
    if '$' in s:
        return s

    s = latex_str(s)

    s2 = s
    for macro in (r'\hspace', r'\vspace', r'\ ', r'\,', r'\;', r'\!',
                  r'\tiny', r'\scriptsize', r'\footnotesize', r'\small',
                  r'\normalsize', r'\large', '\Large',
                  r'\LARGE', r'\huge', r'\Huge'):
        s2 = s2.replace(macro, '')

    # If have _, ^, \frac, etc.  need to be in math-mode.  Should
    # prevent lcapy generating such strings and warn user to explicity
    # use math mode.  The tricky part is that arbitrary signals may
    # have math symbols, say sqrt.
    math_symbols = ('_',  '^', '\\')
    for math_symbol in math_symbols:
        if math_symbol in s2:
            return '$' + s + '$'
    return s


def latex_format_node_label(s):

    if s[0] == '_':
        return s.replace('_', r'\_')

    parts = s.split('_')
    if len(parts) > 2:
        raise ValueError('Double subscript in %s' % s)
    elif len(parts) == 2:
        return '%s$_{\mathrm{%s}}$' % (parts[0], parts[1])

    return s



from lcapy import *

a = Circuit('VRC1.sch')

v = a.R.v.subs(omega0, 3)

ax = v.plot((-1, 10))

from matplotlib.pyplot import savefig
savefig(__file__.replace('.py', '.png'), bbox_inches='tight')








"""This module provides difference equation support.

Copyright 2021 Michael Hayes, UCECE

"""

from .expr import equation
from .nexpr import DiscreteTimeDomainExpression, nexpr
import sympy as sym


class DifferenceEquation(DiscreteTimeDomainExpression):

    def __init__(self, lhs, rhs, inputsym='x', outputsym='y', **assumptions):

        lhs = nexpr(lhs, **assumptions)
        rhs = nexpr(rhs, **assumptions)

        lhssymbols = lhs.symbols
        rhssymbols = rhs.symbols

        if inputsym not in rhssymbols:
            raise ValueError('Input symbol %s not in rhs %s' % (inputsym, rhs))
        if outputsym not in lhssymbols:
            raise ValueError('Output symbol %s not in lhs %s' %
                             (outputsym, lhs))

        cls = lhs.__class__
        eqn = sym.Eq(lhs.expr, rhs.expr, evaluate=False)
        super(DifferenceEquation, self).__init__(cls(eqn), **assumptions)
        self.inputsym = inputsym
        self.outputsym = outputsym

    def __eq__(self, x):

        if not isinstance(x, self.__class__):
            return False

        return self.lhs == x.lhs and self.rhs == x.rhs

    def transfer_function(self):
        """Create discrete-time transfer function."""
        from .zexpr import zexpr

        X = self.inputsym.upper()
        Y = self.outputsym.upper()

        result = self.ZT().solve(Y)[0] / zexpr(X + '(z)')
        result.is_causal = True

        return result

    @property
    def lhs(self):
        return DiscreteTimeDomainExpression(self.expr.lhs, **self.assumptions)

    @property
    def rhs(self):
        return DiscreteTimeDomainExpression(self.expr.rhs, **self.assumptions)

    def dlti_filter(self):
        """Create discrete-time linear time-invariant filter."""

        return self.transfer_function().dlti_filter()

    def separate(self):
        """Rewrite difference equation so that input symbols are on the right
        and output symbols are on the left."""

        newlhs = 0
        newrhs = 0

        for term in self.lhs.as_ordered_terms():
            symbols = nexpr(term).symbols
            if self.outputsym in symbols:
                newlhs += term
            if self.inputsym in symbols:
                newrhs -= term

        for term in self.rhs.as_ordered_terms():
            symbols = nexpr(term).symbols
            if self.outputsym in symbols:
                newlhs -= term
            if self.inputsym in symbols:
                newrhs += term

        return self.__class__(newlhs, newrhs, **self.assumptions)



from lcapy import (texpr, cos, sin, exp, delta, sign, t, H, ramp,
                   rampstep, rect, tri, symbol, s, pi, j, Derivative,
                   Integral)

alpha = symbol('alpha')
t0 = symbol('t0')
f0 = symbol('f0')
w0 = 2 * pi * f0

sigs = [texpr('x(t)'), texpr('x(a * t)'), texpr('x(t - tau)'),
        cos(w0 * t), sin(w0 * t), exp(j * w0 * t),
        texpr(1), t, t**2, delta(t), delta(t - t0),
        H(t), t * H(t), sign(t),
        exp(-abs(t)), exp(-t) * H(t),
        rect(t - 0.5), tri(t - 1), ramp(t), rampstep(t),
        Derivative('x(t)', t), Derivative('x(t)', t, 2),
        Integral('x(t)', (t, 0, t))]

for sig in sigs:
    print(':math:`%s \\longleftrightarrow %s`\n' %
          (sig.latex(), sig(s).latex()))



"""This module performs parsing of SPICE-like netlists.  It uses a
custom parser rather than lex/yacc to give better error messages.

Copyright 2015--2022 Michael Hayes, UCECE

"""

import re

# Could use a script to generate parser and parsing tables if speed
# was important.

# Each line of a netlist (called net here) has a name followed by a
# number of required parameters (params), a number of optional
# parameters, an optional semicolon, and optional comma separated
# key-value pairs.
#
# Each param can be a node, keyword, name, or value.
#
# Here's an example: I1 4 5 {Piecewise((V/L, t >= 0))}; right
# The name is I1; 4, 5 are nodes; the expression is brackets is a value.
#
# The name or value params are referred to as args.


def split(s, delimiters):
    """Split string by specified delimiters but not if a delimiter is
    within curly brackets {} or ""."""

    parts = []
    current = []
    close_bracket = ''
    bracket_stack = []
    for c in (s + delimiters[0]):
        if c in delimiters and len(bracket_stack) == 0:
            if len(current) > 0:
                parts.append(''.join(current))
            current = []
        else:
            if c == close_bracket:
                close_bracket = bracket_stack.pop()
            elif c == '{':
                bracket_stack.append(close_bracket)
                close_bracket = '}'
            elif c == '"':
                bracket_stack.append(close_bracket)
                close_bracket = '"'
            current.append(c)
    if close_bracket != '':
        raise ValueError('Missing %s in %s' % (close_bracket, s))
    return parts


class ParamDef:

    def __init__(self, name, base, comment):

        self.name = name
        self.base = base
        self.comment = comment
        self.baseclass = None

    def is_valid(self, string):

        if self.baseclass is None:
            return True
        return self.baseclass.is_valid(string)


class Param:

    def __init__(self, paramstr, paramdict):

        self.optional = paramstr[0] == '['

        if self.optional:
            paramstr = paramstr[1:-1]

        parts = paramstr.split('=')
        self.name = parts[0]

        if self.name not in paramdict:
            raise ValueError('Unknown parameter ' + self.name)
        self.kind = paramdict[self.name].base

        if len(parts) == 1:
            self.default = None
        else:
            self.default = parts[1]

        self.lowercase_name = self.name.lower()


class Arg:

    def __init__(self, param, name):
        """Set default value."""

        self.name = param.name
        self.assigned = False

        if param.default == 'name':
            default = name
        else:
            default = param.default
        self.value = default

    def assign(self, value):

        if self.assigned:
            raise ValueError('Param %s already assigned' % self.name)

        if value[0] in '{"':
            value = value[1:-1]
        self.value = value
        self.assigned = True


class Args(list):

    def index(self, name):

        name = name.lower()
        for m, arg in enumerate(self):
            if arg.name.lower() == name:
                return m
        return -1


class Rule:

    def __init__(self, cpt_type, classname, fields, params, comment, pos):

        self.type = cpt_type
        self.classname = classname
        self.fields = fields
        self.params = params
        self.comment = comment
        self.pos = pos

    def __repr__(self):

        return self.type + 'name ' + ' '.join(self.fields[1:])

    def syntax_error(self, error, string):

        raise ValueError('Syntax error: %s when parsing %s\nExpected format: %s' % (
            error, string, repr(self)))

    def extract_nodes(self, string, fields, name, namespace):

        ruleparams = self.params
        nodes = []

        for m, ruleparam in enumerate(ruleparams):

            if ruleparam.kind not in ('pin', 'node'):
                continue

            if m >= len(fields):
                self.syntax_error('Missing node %s' % ruleparam.name, string)

            field = fields[m]
            if field[0] == '.':
                # Note, name contains namespace
                field = name + field
            else:
                field = namespace + field
            nodes.append(field)

        return tuple(nodes)

    def extract_args(self, string, fields, name, namespace, default_value):

        args = Args()
        m2 = 0
        for m, ruleparam in enumerate(self.params):

            if ruleparam.kind not in ('name', 'value'):
                m2 = m + 1
                continue

            if m >= len(fields):
                if not ruleparam.optional:
                    self.syntax_error('Missing arg %s' % ruleparam, string)

            args.append(Arg(ruleparam, default_value))

        # Handle unnamed params
        m = 0
        for field in fields[m2:]:
            parts = split(field, '=')
            if len(parts) > 1:
                break
            args[m].assign(field)
            m += 1

        # Handle named params
        for field in fields[m2 + m:]:
            parts = split(field, '=')
            if len(parts) < 2:
                self.syntax_error(
                    'Cannot have value %s after named param' % field, string)
            parts = split(field, '=')

            index = args.index(parts[0])
            if index < 0:
                self.syntax_error('Unknown param ' + parts[0], string)

            args[index].assign(parts[1])

        args = [arg.value for arg in args]

        return args

    def process(self, string, fields, name, namespace, default_value):

        ruleparams = self.params
        if len(fields) > len(ruleparams):
            extra = ''
            if '(' in string:
                extra = ' (perhaps enclose expressions with parentheses in {})'
            self.syntax_error('Too many args' + extra, string)

        nodes = self.extract_nodes(string, fields, name, namespace)

        args = self.extract_args(string, fields, name,
                                 namespace, default_value)

        return nodes, args


class Parser:

    def __init__(self, cpts, grammar, allow_anon=False):
        """cpts is a module containing a class for each component
        grammar is a module defining the syntax of a netlist"""

        # A string defining the syntax for a netlist
        rules = grammar.rules
        # A string defining parameters
        params = grammar.params
        # A string defining delimiter characters
        self.delimiters = grammar.delimiters
        # A string defining comment characters
        self.comments = grammar.comments
        self.allow_anon = allow_anon

        self.cpts = cpts
        self.paramdict = {}
        self.ruledict = {}

        for param in params.split('\n'):
            self._add_param(param)

        for rule in rules.split('\n'):
            self._add_rule(rule)

        cpts = sorted(self.ruledict.keys(), key=len, reverse=True)

        # The symbol name must be a valid Sympy symbol name so
        # it cannot include symbols such as + and -.
        self.cpt_pattern = re.compile("(%s)([#_\w'?]+)?" % '|'.join(cpts))

    def _add_param(self, string):

        if string == '':
            return

        fields = string.split(':')
        paramname = fields[0]
        fields = fields[1].split(';', 1)
        parambase = fields[0].strip()
        comment = fields[1].strip()

        self.paramdict[paramname] = ParamDef(paramname, parambase, comment)

    def _add_rule(self, string):

        if string == '':
            return

        fields = string.split(':')
        cpt_classname = fields[0]
        fields = fields[1].split(';', 1)
        string = fields[0].strip()
        comment = fields[1].strip()

        fields = string.split(' ')

        # Skip the name part in the rule, e.g., only consider D from Dname.
        cpt_type = fields[0][0:-4]

        pos = None
        params = []
        for m, paramstr in enumerate(fields[1:]):

            param = Param(paramstr, self.paramdict)
            params.append(param)
            if pos is None and param.kind == 'keyword':
                pos = m

        if cpt_type not in self.ruledict:
            self.ruledict[cpt_type] = ()
        self.ruledict[cpt_type] += (Rule(cpt_type, cpt_classname, fields,
                                         params, comment, pos), )

    def parse(self, string, namespace='', parent=None):
        """Parse string and create object"""

        directive = False
        net = string.strip()
        if net == '':
            directive = True
        elif net[0] in self.comments:
            directive = True
        elif net[0] == ';':
            directive = True
        elif net[0] == '.':
            directive = True

        if directive:
            cpt_type = 'XX'
            cpt_id = ''
            name = 'XX'
            name += parent._make_anon_cpt_id(cpt_type)
            defname = namespace + cpt_type + cpt_id

            if string.startswith(';') and not string.startswith(';;'):
                opts_string = string[1:]
            else:
                opts_string = ''

            return self.cpts.make('XX', parent, '', defname, name,
                                  cpt_type, cpt_id, string, opts_string, (), '',
                                  Args())

        net = namespace + net
        parts = net.split(';', 1)

        fields = split(parts[0], self.delimiters)

        name = fields.pop(0)
        parts = name.split('.')
        namespace = ''
        if len(parts) > 1:
            namespace = '.'.join(parts[0:-1]) + '.'
        name = parts[-1]

        match = self.cpt_pattern.match(name)
        if match is None:
            raise ValueError(
                'Unknown component %s while parsing "%s"' % (name, net))

        groups = match.groups()
        cpt_type, cpt_id = groups[0], groups[1]
        if cpt_id is None:
            cpt_id = ''

        # This is the most hackery aspect of this parser where we
        # choose the rule pattern based on a keyword.  If the
        # keyword is not present, default to first rule pattern.
        # Perhaps a factory should sort this out?
        rule = self.ruledict[cpt_type][0]
        keyword = ''

        for rule1 in self.ruledict[cpt_type]:
            pos = rule1.pos
            if pos is None:
                continue
            if len(fields) > pos and fields[pos].lower() == rule1.params[pos].lowercase_name:
                rule = rule1
                keyword = rule1.params[pos].name
                break

        defname = namespace + cpt_type + cpt_id
        name = defname
        if (cpt_id == '' and parent is not None
                and (cpt_type in ('A', 'W', 'O', 'P')) or self.allow_anon):
            name += parent._make_anon_cpt_id(cpt_type)
        elif cpt_id == '?':
            # Automatically name cpts to ensure they are unique
            name = name[:-1] + parent._make_anon_cpt_id(cpt_type)

        default_value = cpt_type + cpt_id
        nodes, args = rule.process(net, fields, name, namespace, default_value)

        parts = net.split(';', 1)
        opts_string = parts[1].strip() if len(parts) > 1 else ''

        keyword = (pos, keyword)

        # self.cpts is either the mnacpts or schemcpts module
        return self.cpts.make(rule.classname, parent, namespace,
                              defname, name, cpt_type, cpt_id, net,
                              opts_string, tuple(nodes), keyword,
                              *args)



"""This module provides common utility functions.

Copyright 2021--2022 Michael Hayes, UCECE
"""

import sympy as sym


def factor_const(expr, var):
    """Extract constant factor from expression and return tuple
    of constant and the rest of the expression.

    For example `a * r(var)` returns `a, r(var)`.

    If have a polynomial expression, the leading coefficient is
    returned as the constant, for example: `2 * s + 4` returns `2, s + 2`.

    """

    # Perhaps use expr.as_coeff_Mul() ?

    if expr.is_polynomial():
        poly = sym.Poly(expr, var)
        const = poly.LC()
        if const == 0 or const == 1:
            return 1, expr
        return const, expr / const

    rest = sym.S.One
    const = sym.S.One
    for factor in expr.as_ordered_factors():
        # Cannot use factor.is_constant() since SymPy 1.2, 1.3
        # barfs for Heaviside(t) and DiracDelta(t)
        if not factor.has(var):
            const *= factor
        else:
            rest *= factor
    return const, rest


def term_const(expr, var):
    """Extract constant term from expression and return tuple
    of constant and the rest of the expression."""

    rest = sym.S.One
    const = sym.S.Zero
    for term in expr.as_ordered_terms():
        # Cannot use factor.is_constant() since SymPy 1.2, 1.3
        # barfs for Heaviside(t) and DiracDelta(t)
        if not term.has(var):
            const += term
        else:
            rest += term
    return const, rest


def scale_shift(expr, var):

    if expr == var:
        return sym.S.One, sym.S.Zero

    expr = expr.expand()
    if not expr.as_poly(var).is_linear:
        raise ValueError(
            'Expression not a linear function of %s: %s' % (var, expr))

    scale = expr.coeff(var, 1)
    shift = expr.coeff(var, 0)

    return scale, shift


def as_N_D(expr, var, monic_denominator=False, use_sympy=False):

    if use_sympy:
        return expr.as_numer_denom()

    factors = expr.as_ordered_factors()

    numers = []
    denoms = []
    for factor in factors:
        if factor.is_Function and factor.func == sym.exp:
            # SymPy treats exp(-s * a) as 1 / exp(s * a)
            numer = factor
            denom = sym.S.One
        else:
            numer, denom = factor.as_numer_denom()
        numers.append(numer)
        denoms.append(denom)

    poly_denom = False
    for denom in denoms:
        if denom != 1 and denom.is_polynomial(var):
            poly_denom = True
            break

    if not poly_denom:
        return expr.as_numer_denom()

    N = sym.S.One
    D = sym.S.One

    for numer, denom in zip(numers, denoms):
        N *= numer
        if denom.is_polynomial(var):
            D *= denom
        else:
            N /= denom

    N = N.simplify()

    if monic_denominator:
        Dpoly = sym.Poly(D, var)
        LC = Dpoly.LC()
        D = Dpoly.monic().as_expr()
        N = (N / LC).simplify()

    return N, D


def as_sum_terms(expr, var):

    N, D = as_N_D(expr, var)
    N = N.simplify()

    return [term / D for term in N.expand().as_ordered_terms()]


def as_sum(expr, var):

    result = 0
    for term in as_sum_terms(expr, var):
        result += term
    return result


def merge_common(lists):
    # From www.geeksforgeeks.org

    from collections import defaultdict

    neighbours = defaultdict(set)
    visited = set()
    for each in lists:
        for item in each:
            neighbours[item].update(each)

    def comp(node, neighbours=neighbours, visited=visited, visit=visited.add):

        nodes = set([node])
        next_node = nodes.pop
        while nodes:
            node = next_node()
            visit(node)
            nodes |= neighbours[node] - visited
            yield node

    for node in neighbours:
        if node not in visited:
            yield sorted(comp(node))


def isiterable(arg):

    return hasattr(arg, '__iter__')


def factor_expr(expr, factor):
    """Extract factor from expression or None if expression does
    not have factor."""

    factors = expr.as_ordered_factors()
    if factor not in factors:
        return None

    return expr / factor


def separate_dirac_delta(expr):

    terms = expr.as_ordered_terms()
    deltas = []
    rest = 0

    for term in terms:
        if term.has(sym.DiracDelta):
            deltas.append(term)
        else:
            rest += term

    return rest, deltas


def split_dirac_delta(expr):
    """Return expression as a list of terms.
    The first term has no DiracDeltas, the second term collates
    the DiracDeltas, the third term collates derivatives of DiracDeltas, etc.

    For example, u(t) + DiractDelta(t, 1) returns [u(t), 0, DiracDelta(t, 1)]
    """

    terms = expr.as_ordered_terms()
    parts = {}
    rest = 0

    # FIXME, DiracDelta needs to be a factor

    for term in terms:
        if term.has(sym.DiracDelta):
            if len(term.args) == 1:
                if 1 not in parts:
                    parts[1] = 0
                parts[1] += term
            else:
                idx = term.args[1] + 1
                if idx not in parts:
                    parts[idx] = 0
                parts[idx] += term
        else:
            parts[0] = term

    maxkey = max(parts.keys())
    result = []
    for key in range(maxkey + 1):
        if key in parts:
            result.append(parts[key])
        else:
            result.append(0)

    return result


def remove_images(expr, var, dt, m1=0, m2=0):

    if m2 == 0 and isinstance(m1, tuple) and len(m1) == 2:
        # Perhaps should warn that this might be deprecated?
        m1, m2 = m1

    remove_all = m1 == 0 and m2 == 0

    const, expr1 = factor_const(expr, var)

    result = sym.S.Zero
    terms = expr1.as_ordered_terms()

    if len(terms) > 1:
        for term in expr1.as_ordered_terms():
            result += remove_images(term, var, dt, m1, m2)
        return const * result

    if not isinstance(expr1, sym.Sum):
        return expr

    sumsym = expr1.args[1].args[0]

    def query(expr):

        return expr.is_Add and expr.has(var) and expr.has(sumsym)

    def value(expr):
        if not expr.is_Add:
            return expr

        if not expr.is_polynomial(var) and not expr.as_poly(var).is_linear:
            return expr
        expr = expr.expand()
        a = expr.coeff(var, 1)
        b = expr.coeff(var, 0)

        if a == 0:
            return expr
        if b / a != -sumsym / dt:
            return expr
        return a * var

    expr1 = expr1.replace(query, value)

    if remove_all:
        return const * expr1.args[0]

    return const * sym.Sum(expr1.args[0], (sumsym, m1, m2))


def pair_conjugates(poles_dict):
    """Return dictionary of conjugate pole pairs and a dictionary of the
    remaining single poles."""

    pole_single_dict = poles_dict.copy()
    pole_pair_dict = {}

    pole_list = list(poles_dict)

    for i, pole in enumerate(pole_list):
        pole_c = sym.conjugate(pole)
        # Check for conjugate pole
        if pole_c in pole_list[i + 1:]:
            pole_single_dict.pop(pole, None)
            pole_single_dict.pop(pole_c, None)

            o1 = poles_dict[pole]
            o2 = poles_dict[pole_c]
            if o1 == o2:
                pole_pair_dict[pole, pole_c] = o1
            elif o1 > o2:
                pole_pair_dict[pole, pole_c] = o2
                pole_single_dict[pole] = o1 - o2
            else:
                pole_pair_dict[pole, pole_c] = o1
                pole_single_dict[pole_c] = o2 - o1

    return pole_pair_dict, pole_single_dict


def similarity_shift(expr, var):
    """Rewrite foo(a * t + b) as foo(t) and return a, b."""

    scale = None
    shift = None
    fail = False

    for expr1 in sym.preorder_traversal(expr):

        if not expr1.is_Function:
            continue

        arg = expr1.args[0]
        if not arg.has(var):
            continue

        poly = arg.as_poly(var)
        if poly is None or not poly.is_linear:
            fail = True
            break

        scale1 = arg.coeff(var, 1)
        shift1 = arg.coeff(var, 0)

        if scale is None:
            scale = scale1
        if shift is None:
            shift = shift1

        if scale != scale1 or shift != shift1:
            fail = True
            break

    if fail or scale is None:
        return expr, 1, 0

    expr2 = expr.replace(var * scale + shift, var)

    return expr2, scale, shift


def expand_functions(expr, var):

    # Try rewriting functions such as rampstep(t - 1)
    # to ramp(t - 1) - ramp(t - 2) to sums of weighted Heavisides...

    const, expr = factor_const(expr, var)

    if expr.is_Function:
        new_expr = expr.rewrite()
        if expr != new_expr:
            return const * expand_functions(new_expr, var)
        return const * new_expr

    terms = expr.as_ordered_terms()
    if len(terms) == 1:
        return const * expr

    expr = 0
    for term in terms:
        expr += expand_functions(term, var)

    return const * expr



"""This module exports all latin and greek letters as Symbols, so you can
conveniently do

    >>> from lcapy.abc import x, y

instead of the slightly more clunky-looking

    >>> from lcapy import symbols
    >>> x, y = symbols('x y')

Unlike sympy.abc, this does not define symbols that clash with Lcapy's
domain variables.

"""

from lcapy.expr import symbols

a, b, c, d, e, g, h, i = symbols('a, b, c, d, e, g, h, i')
l, m, o, p, q, r = symbols('l, m, o, p, q, r')
u, v, w, x, y = symbols('u, v, w, x, y')

A, B, C, D, E, G, H, I, J = symbols('A, B, C, D, E, G, H, I, J')
K, L, M, N, O, P, Q, R, S, T = symbols('K, L, M, N, O, P, Q, R, S, T')
U, V, W, X, Y, Z = symbols('U, V, W, X, Y, Z')

alpha, beta, gamma, delta = symbols('alpha, beta, gamma, delta')
epsilon, zeta, eta, theta = symbols('epsilon, zeta, eta, theta')
iota, kappa, lamda, mu = symbols('iota, kappa, lamda, mu')
nu, xi, omicron, pi = symbols('nu, xi, omicron, pi')
rho, sigma, tau, upsilon = symbols('rho, sigma, tau, upsilon')
phi, chi, psi = symbols('phi, chi, psi')



"""
This module performs plotting using matplotlib.

Copyright 2014--2022 Michael Hayes, UCECE
"""

import numpy as np
from .symbols import pi
from .utils import separate_dirac_delta, factor_const
from sympy import DiracDelta, solve
from warnings import warn


# Perhaps add Formatter classes that will produce the plot data?


def make_axes(figsize=None, axes=None, **kwargs):

    from matplotlib.pyplot import subplots

    if axes is not None:
        if isinstance(axes, tuple):
            # FIXME
            axes = axes[0]
        fig = axes.figure
    elif figsize is not None:
        fig, axes = subplots(1, figsize=figsize, **kwargs)
    else:
        fig, axes = subplots(1, **kwargs)

    return axes


def parse_range(frange, zero=0.1, positive=False):

    # Ensure that have floats.  This deals with SymPy symbols.
    fmin = float(frange[0])
    fmax = float(frange[1])

    if fmax < fmin:
        raise ValueError('Max value smaller than min value')

    if positive and fmin < 0:
        raise ValueError('Min value not positive.')

    if positive and fmin == 0:
        fmin = zero

    return fmin, fmax


def plot_deltas(ax, t, deltas, var, plot_type='real', color='k'):

    vals = []
    for delta in deltas:
        delta = delta.expand(diracdelta=True, wrt=var)
        const, expr = factor_const(delta, var)
        if not (expr.is_Function and expr.func is DiracDelta):
            raise ValueError('Cannot handle %s' % expr)
        t0 = solve(expr.args[0], var)[0]

        const = complex(const)
        if plot_type in ('real', 'default'):
            const = const.real
        elif plot_type == 'imag':
            const = const.imag
        elif plot_type == 'abs':
            const = abs(const)
        elif plot_type == 'dB':
            const = 20 * np.log10(abs(const))
        elif plot_type == 'phase':
            const = np.angle(const)
        elif plot_type == 'radians':
            const = np.angle(const)
        elif plot_type == 'degrees':
            const = np.degrees(np.angle(const))
        else:
            raise ValueError('Unhandled plot type %s' % plot_type)

        if t0 >= min(t) and t0 <= max(t):
            vals.append((t0, const))

    vals = np.array(vals)

    tmin, tmax = ax.get_xlim()
    T = tmax - tmin
    ymin, ymax = ax.get_ylim()
    yvalsmax = vals[:, 1].max()

    for t0, y in vals:

        # TODO: determine best scaling
        y1 = y / yvalsmax * ymax / 2

        st = T / 10
        sy = ymax

        ax.arrow(t0, 0, 0, y1, fc=color, ec=color,
                 width=0.025 * st, head_width=0.1 * st,
                 head_length=0.1 * sy, overhang=0,
                 length_includes_head=True, clip_on=False)

        # TODO: nicely format the scale factor
        ax.text(t0 + T / 100, y1 + ymax / 100, '%.*g' % (3, y))


def plot_pole_zero(obj, **kwargs):

    from matplotlib.pyplot import Circle, rcParams

    obj = obj.doit()

    poles = obj.poles()
    zeros = obj.zeros()
    try:
        p = np.array([p.cval for p in poles.keys()])
        z = np.array([z.cval for z in zeros.keys()])
    except TypeError:
        raise TypeError('Cannot plot poles and zeros of symbolic expression')

    ax = make_axes(figsize=kwargs.pop('figsize', None),
                   axes=kwargs.pop('axes', None))

    unitcircle = kwargs.pop('unitcircle', False)

    ax.axvline(0, color='0.7')
    ax.axhline(0, color='0.7')

    a = np.hstack((p, z))

    if unitcircle:
        ax.add_artist(Circle((0, 0), 1, color='blue',
                      linestyle='--', fill=False))

    x_min = a.real.min()
    x_max = a.real.max()
    y_min = a.imag.min()
    y_max = a.imag.max()

    if unitcircle:
        if x_min > -1:
            x_min = -1
        if x_max < 1:
            x_max = 1
        if y_min > -1:
            y_min = -1
        if y_max < 1:
            y_max = 1

    x_extra, y_extra = 0.0, 0.0

    # This needs tweaking for better bounds.
    if len(a) >= 2:
        x_extra, y_extra = 0.1 * (x_max - x_min), 0.1 * (y_max - y_min)
    if x_extra == 0:
        x_extra += 1.0
    if y_extra == 0:
        y_extra += 1.0

    x_min -= 0.5 * x_extra
    x_max += 0.5 * x_extra
    if unitcircle:
        bbox = ax.get_window_extent()
        aspect = bbox.width / bbox.height

        x_min *= aspect
        x_max *= aspect

    ax.axis('equal')
    ax.set_xlim(x_min, x_max)
    # overconstrained so ignored
    #ax.set_ylim(y_min - 0.5 * y_extra, y_max + 0.5 * y_extra)

    def annotate(axes, poles, offset=None):
        if offset is None:
            xmin, xmax = axes.get_xlim()
            offset = (xmax - xmin) / 40

        for pole, num in poles.items():
            if num > 1:
                p = pole.cval
                axes.text(p.real + offset, p.imag + offset, '%d' % num)

    # Marker size.  Unfortunately, the default is too small
    # but if the user wants a size of 6 set in rcParams, they are out of luck...
    if 'ms' not in kwargs and 'markersize' not in kwargs and rcParams['lines.markersize'] == 6:
        kwargs['ms'] = 19

    fillstyle = kwargs.pop('fillstyle', 'none')
    xlabel = kwargs.pop('xlabel', 'Re(%s)' % obj.var)
    ylabel = kwargs.pop('ylabel', 'Im(%s)' % obj.var)
    title = kwargs.pop('title', None)

    lines = ax.plot(z.real, z.imag, 'o', fillstyle=fillstyle, **kwargs)
    annotate(ax, zeros)
    color = kwargs.pop('color', lines[0].get_color())
    ax.plot(p.real, p.imag, 'x', fillstyle=fillstyle, color=color, **kwargs)
    annotate(ax, poles)

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    ax.grid(True)

    return ax


def plotit(ax, obj, f, V, plot_type=None, deltas=None, log_magnitude=False,
           log_frequency=False, norm=False, dbmin=-120, unwrap=False, **kwargs):

    plots = {(True, True): ax.loglog,
             (True, False): ax.semilogy,
             (False, True): ax.semilogx,
             (False, False): ax.plot}

    label = ''
    if hasattr(obj, 'quantity_label'):
        label = obj.quantity_label
    part = obj.part
    units = str(obj.units)

    if plot_type == 'real':
        V = V.real
        part = 'real part'
    elif plot_type == 'imag':
        V = V.imag
        part = 'imag part'
    elif plot_type == 'magnitude':
        V = abs(V)
        part = 'magnitude'
    elif plot_type == 'dB':
        Vabs = abs(V)
        Vmin = 10**(dbmin / 20)
        if np.all(Vabs < Vmin):
            warn('All values below dbmin')
        Vabs[Vabs < Vmin] = Vmin
        V = 20 * np.log10(Vabs)
        part = 'magnitude'
        units = 'dB'
    elif plot_type == 'radians':
        V = np.angle(V)
        if unwrap:
            V = np.unwrap(V)
        part = 'phase'
        units = 'radians'
    elif plot_type == 'degrees':
        V = np.angle(V)
        if unwrap:
            V = np.unwrap(V)
        V = np.degrees(V)
        part = 'phase'
        units = 'degrees'

    if obj.is_magnitude or np.all(V > 0):
        plot = plots[(log_magnitude, log_frequency)]
    else:
        plot = plots[(False, log_frequency)]

    if norm:
        default_xlabel = 'Normalized frequency'
    else:
        default_xlabel = obj.domain_label_with_units

    if label == '':
        default_ylabel = part.capitalize()
    else:
        default_ylabel = label + ' ' + part
    if units != '' and units != '1':
        default_ylabel += ' (' + units + ')'

    xlabel = kwargs.pop('xlabel', default_xlabel)
    ylabel = kwargs.pop('ylabel', default_ylabel)
    ylabel2 = kwargs.pop('ylabel2', default_ylabel)
    second = kwargs.pop('second', False)
    xscale = kwargs.pop('xscale', 1)
    yscale = kwargs.pop('yscale', 1)
    title = kwargs.pop('title', None)

    lines = plot(f * xscale, V * yscale, **kwargs)
    color = kwargs.pop('color', lines[0].get_color())

    if deltas is not None:
        plot_deltas(ax, f * xscale, deltas, obj.var, plot_type, color)

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    ylabel = ylabel2 if second else ylabel
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    ax.grid(True)
    return lines


def plot_frequency(obj, f, plot_type=None, **kwargs):

    obj = obj.doit()

    # Much of the hoop jumping is to speed up plotting since
    # obj.real can be slow.  Instead we evaluate complex
    # objects and then convert to phase, magnitude, etc.

    norm = kwargs.pop('norm', False)
    npoints = kwargs.pop('npoints', 400)
    log_magnitude = kwargs.pop('log_magnitude', False)
    log_frequency = kwargs.pop(
        'log_frequency', False) or kwargs.pop('log_scale', False)
    if kwargs.pop('loglog', False):
        log_magnitude = True
        log_frequency = True

    dbmin = kwargs.pop('dbmin', -120)
    label = kwargs.pop('label', None)
    nyticks = kwargs.pop('nyticks', None)
    color2 = kwargs.pop('color2', None)
    linestyle2 = kwargs.pop('linestyle2', '--')
    unwrap = kwargs.pop('unwrap', False)

    # FIXME, determine useful frequency range...
    if f is None:
        if norm:
            f = (-0.5, 0.5)
        else:
            f = (0, 2)
    if isinstance(f, (int, float)):
        f = (0, f)
    if isinstance(f, tuple):
        if log_frequency:
            fmin, fmax = parse_range(f, 1e-1, positive=True)
            f = np.geomspace(fmin, fmax, npoints)
        else:
            fmin, fmax = parse_range(f, 1e-1, positive=False)
            f = np.linspace(fmin, fmax, npoints)

    # Objects can have a `part` attribute that is set by methods such
    # as real, imag, phase, magnitude.  If this is defined,
    # `plot_type` is ignored.

    plot1_type = 'default'
    plot2_type = None

    V = obj.evaluate(f)

    types = ['dB-phase', 'dB-radians', 'dB-phase-degrees',
             'dB-degrees', 'mag-phase', 'magnitude-phase',
             'mag-phase-degrees', 'magnitude-phase-degrees',
             'real-imag', 'mag', 'magnitude', 'phase', 'radians',
             'phase-degrees', 'degrees', 'real', 'imag', 'dB', 'abs']
    if plot_type is not None and plot_type not in types:
        raise ValueError('Unknown plot type %s, expecting: %s ' %
                         (plot_type, ', '.join(types)))

    if not V.dtype == complex:
        if plot_type is None:
            plot_type = 'real'

    elif obj.part == '':
        if plot_type is None:
            plot_type = 'dB-phase'

    if plot_type in ('dB_phase', 'dB-phase', 'dB-radians'):
        plot1_type = 'dB'
        plot2_type = 'radians'
    elif plot_type in ('dB_phase_degrees', 'dB-phase-degrees',
                       'dB-degrees'):
        plot1_type = 'dB'
        plot2_type = 'degrees'
    elif plot_type in ('mag_phase', 'magnitude_phase',
                       'mag-phase', 'magnitude-phase'):
        plot1_type = 'magnitude'
        plot2_type = 'radians'
    elif plot_type in ('mag_phase_degrees', 'magnitude_phase_degrees',
                       'mag-phase-degrees', 'magnitude-phase-degrees'):
        plot1_type = 'magnitude'
        plot2_type = 'degrees'
    elif plot_type in ('real_imag', 'real-imag'):
        plot1_type = 'real'
        plot2_type = 'imag'
    elif plot_type in ('mag', 'magnitude'):
        plot1_type = 'magnitude'
    elif plot_type in ('phase', 'radians'):
        plot1_type = 'radians'
    elif plot_type in ('phase-degrees', 'degrees'):
        plot1_type = 'degrees'
    elif plot_type == 'abs':
        plot1_type = 'abs'
    elif plot_type == 'real':
        plot1_type = 'real'
    elif plot_type == 'imag':
        plot1_type = 'imag'
    elif plot_type == 'dB':
        plot1_type = 'dB'
    else:
        raise RuntimeError('Internal error')

    deltas = None
    if kwargs.pop('plot_deltas', True) and obj.has(DiracDelta):
        cls = obj.__class__
        rest, deltas = separate_dirac_delta(obj.expr)
        obj = cls(rest, **obj.assumptions)
        V = obj.evaluate(f)

    ax = make_axes(figsize=kwargs.pop('figsize', None),
                   axes=kwargs.pop('axes', None))

    if label is None:
        label = plot1_type

    lines = plotit(ax, obj, f, V, plot1_type, deltas,
                   log_frequency=log_frequency,
                   log_magnitude=log_magnitude, norm=norm,
                   label=label, dbmin=dbmin, unwrap=unwrap, **kwargs)

    if plot2_type is None:
        return ax

    if color2 is None:
        color2 = lines[0].get_color()

    kwargs2 = kwargs.copy()
    for key in ('color', 'linestyle', 'xlabel', 'xlabel2', 'ylabel', 'title', 'xscale', 'yscale'):
        kwargs2.pop(key, None)

    # Dummy plot to add label to legend.
    ax.plot([], [], label=plot2_type, color=color2,
            linestyle=linestyle2, **kwargs2)

    ax2 = ax.twinx()
    kwargs['axes'] = ax2
    kwargs.pop('color', None)

    plotit(ax2, obj, f, V, plot2_type, deltas, log_frequency=log_frequency,
           log_magnitude=log_magnitude, norm=norm, second=True, color=color2,
           linestyle=linestyle2, unwrap=unwrap, **kwargs)

    if plot2_type in ('phase', 'radians') and not unwrap:
        ax2.set_ylim(-np.pi, np.pi)
    elif plot2_type in ('phase-degrees', 'degrees') and not unwrap:
        ax2.set_ylim(-180, 180)

    if plot1_type == 'dB':
        Vabs = abs(V)
        Vmin = 10**(dbmin / 20)
        if np.all(Vabs < Vmin):
            warn('All values below dbmin')
        Vabs[Vabs < Vmin] = Vmin

        m = np.isinf(Vabs)
        if any(m):
            warn('Approximating infinite value!')
            Vabs[m] = 1e30

        dB = 20 * np.log10(Vabs)
        dBmin = min(dB)
        dBmax = max(dB)

        ymin = np.floor(dBmin / 10) * 10
        ymax = np.ceil(dBmax / 10) * 10
        yrange = ymax - ymin

        if False:
            # Provide some wiggle room
            ymin -= yrange / 10
            ymax += yrange / 10
            yrange = ymax - ymin

        # This will clobber previous plot values
        ax.set_ylim(ymin, ymax)

        if nyticks is None:
            # An odd number is better to show zero phase.
            for nyticks in (7, 6, 5, 8, 9):
                if yrange / (nyticks - 1) in (20, 10, 5, 2):
                    break

    from matplotlib import ticker

    if nyticks is None:
        nyticks = 5

    ax.yaxis.set_major_locator(ticker.LinearLocator(nyticks))
    ax2.yaxis.set_major_locator(ticker.LinearLocator(nyticks))

    ax.legend()
    return ax, ax2


def plot_bode(obj, f, **kwargs):

    if 'log_frequency' not in kwargs:
        kwargs['log_frequency'] = True
    if 'unwrap' not in kwargs:
        kwargs['unwrap'] = True

    return plot_frequency(obj, f, **kwargs)


def plot_angular_bode(obj, f, **kwargs):

    if 'log_frequency' not in kwargs:
        kwargs['log_frequency'] = True

    return plot_angular_frequency(obj, f, **kwargs)


def plot_angular_frequency(obj, omega, plot_type=None, **kwargs):

    npoints = kwargs.pop('npoints', 400)
    norm = kwargs.pop('norm', False)
    log_frequency = kwargs.pop(
        'log_frequency', False) or kwargs.pop('log_scale', False)

    # FIXME, determine useful frequency range...
    if omega is None:
        if norm:
            omega = (-np.pi, np.pi)
        else:
            omega = (0, np.pi)
    if isinstance(omega, (int, float)):
        omega = (0, omega)
    if isinstance(omega, tuple):
        if log_frequency:
            wmin, wmax = parse_range(omega, 1e-1, positive=True)
            omega = np.geomspace(wmin, wmax, npoints)
        else:
            wmin, wmax = parse_range(omega, 1e-1, positive=False)
            omega = np.linspace(wmin, wmax, npoints)

    if norm and 'xlabel' not in kwargs:
        kwargs['xlabel'] = 'Normalized angular frequency'

    return plot_frequency(obj, omega, plot_type=plot_type,
                          norm=norm, log_frequency=log_frequency, **kwargs)


def plot_time(obj, t, plot_type=None, **kwargs):

    obj = obj.doit()
    npoints = kwargs.pop('npoints', 400)

    # FIXME, determine useful time range...
    if t is None:
        t = (-0.2, 2)
    if isinstance(t, (int, float)):
        t = (0, t)
    if isinstance(t, tuple):
        tmin, tmax = parse_range(t, 1e-1, positive=False)
        t = np.linspace(tmin, tmax, npoints)

    deltas = None
    if kwargs.pop('plot_deltas', True) and obj.has(DiracDelta):
        cls = obj.__class__
        rest, deltas = separate_dirac_delta(obj.expr)
        obj = cls(rest, **obj.assumptions)

    v = obj.evaluate(t)

    if plot_type == None:
        plot_type = 'real'
    if plot_type == 'real':
        v = v.real
    elif plot_type == 'imag':
        v = v.imag
    elif plot_type == 'abs':
        v = abs(v)
    else:
        raise ValueError(
            'Invalid plot_type: must be real, imag, real-imag, abs')

    ax = make_axes(figsize=kwargs.pop('figsize', None),
                   axes=kwargs.pop('axes', None))

    xlabel = kwargs.pop('xlabel', obj.domain_label_with_units)
    ylabel = kwargs.pop('ylabel', obj.label_with_units)
    xscale = kwargs.pop('xscale', 1)
    yscale = kwargs.pop('yscale', 1)
    title = kwargs.pop('title', None)

    lines = ax.plot(t * xscale, v * yscale, **kwargs)
    color = kwargs.pop('color', lines[0].get_color())

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    if deltas is not None:
        # TODO, fix yscale
        plot_deltas(ax, t * xscale, deltas, obj.var, plot_type, color=color)

    ax.grid(True)
    return ax


# make a stem plot for complex values sequences in the complex domain
def plot_sequence_polar(obj, ni=(-10, 10), **kwargs):

    obj = obj.doit()
    npoints = kwargs.pop('npoints', 400)

    # FIXME, determine useful range...
    if ni is None:
        ni = (-20, 20)
    if isinstance(ni, tuple):
        # Use float data type since NumPy barfs for n**(-a) where a is
        # an integer.
        ni = np.arange(ni[0], ni[1] + 1, dtype=float)

    v = obj.evaluate(ni)

    ax = make_axes(figsize=kwargs.pop('figsize', None),
                   axes=kwargs.pop('axes', None),
                   subplot_kw=dict(polar=True))

    phi = np.angle(v)
    mag = abs(v)

    # Plot symbols
    lines = ax.plot(phi, mag, 'o', **kwargs)

    color = kwargs.pop('color', lines[0].get_color())

    # Plot lines from origin
    Nv = len(v)
    ax.plot((phi, phi), (np.zeros(Nv), mag), color=color, **kwargs)
    ax.grid(True)
    return ax


def plot_sequence(obj, ni, plot_type=None, polar=False, **kwargs):

    from matplotlib.ticker import MaxNLocator

    obj = obj.doit()

    if polar:
        return plot_sequence_polar(obj, ni, **kwargs)

    npoints = kwargs.pop('npoints', 400)

    # FIXME, determine useful range...
    if ni is None:
        ni = (-20, 20)
    if isinstance(ni, tuple):
        # Use float data type since NumPy barfs for n**(-a) where a is
        # an integer.
        ni = np.arange(ni[0], ni[1] + 1, dtype=float)

    v = obj.evaluate(ni)

    if plot_type == None:
        plot_type = 'real'
    if plot_type == 'real':
        v = v.real
    elif plot_type == 'imag':
        v = v.imag
    elif plot_type == 'abs':
        v = abs(v)
    else:
        raise ValueError(
            'Invalid plot_type: must be real, imag, real-imag, abs')

    deltas = None
    if obj.has(DiracDelta):
        cls = obj.__class__
        rest, deltas = separate_dirac_delta(obj.expr)
        obj = cls(rest, **obj.assumptions)

    ax = make_axes(figsize=kwargs.pop('figsize', None),
                   axes=kwargs.pop('axes', None))

    xlabel = kwargs.pop('xlabel', obj.domain_label_with_units)
    ylabel = kwargs.pop('ylabel', obj.label_with_units)
    xscale = kwargs.pop('xscale', 1)
    yscale = kwargs.pop('yscale', 1)
    title = kwargs.pop('title', None)
    color = kwargs.pop('color', None)

    markerline, stemlines, baseline = ax.stem(
        ni * xscale, v * yscale, use_line_collection=True, **kwargs)
    if color is not None:
        stemlines.set_color(color)
        markerline.set_color(color)
    else:
        color = stemlines.get_color()

    # Ensure integer ticks.
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    if deltas is not None:
        # TODO, fix yscale
        plot_deltas(ax, ni * xscale, deltas, obj.var, plot_type, color=color)

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    ax.grid(True)
    return ax


def plot_phasor(obj, **kwargs):

    obj = obj.doit()

    ax = make_axes(figsize=kwargs.pop('figsize', None),
                   axes=kwargs.pop('axes', None),
                   subplot_kw=dict(polar=True))

    phi = obj.phase.fval
    mag = obj.magnitude.fval

    ax.plot((phi, phi), (0, mag), **kwargs)
    return ax


def plot_nyquist(obj, f, norm=False, **kwargs):

    from matplotlib.pyplot import Circle, rcParams

    npoints = kwargs.pop('npoints', 400)
    # Default to True to get better plots.
    log_frequency = kwargs.pop('log_frequency', False)
    unitcircle = kwargs.pop('unitcircle', True)

    fn = None

    # FIXME, determine useful frequency range...
    if f is None:
        f = (-100, 100)
    if isinstance(f, (int, float)):
        f = (0, f)
    if isinstance(f, tuple):
        fmin, fmax = parse_range(f, 1e-1, positive=False)
        if log_frequency:
            if fmin < 0:
                frange = fmax - fmin
                f = np.geomspace(frange / 1e4, fmax, npoints // 2)
                f = np.hstack((0, f))
                fn = -np.geomspace(frange / 1e4, abs(fmin), npoints // 2)
                fn = np.hstack((0, fn))
            elif fmin == 0:
                f = np.geomspace(fmax / 1e4, fmax, npoints // 2)
                f = np.hstack((0, f))
            else:
                f = np.geomspace(fmin, fmax, npoints)
        else:
            fmin, fmax = parse_range(f, 1e-1, positive=False)
            f = np.linspace(fmin, fmax, npoints)

    V = obj.evaluate(f)
    if not V.dtype == complex:
        raise ValueError('Data not complex')

    obj = obj.doit()

    ax = make_axes(figsize=kwargs.pop('figsize', None),
                   axes=kwargs.pop('axes', None))

    if unitcircle:
        ax.add_artist(Circle((0, 0), 1, color='blue',
                      linestyle='--', fill=False))

    lines = ax.plot(V.real, V.imag)

    if fn is not None:
        V = obj.evaluate(fn)
        color = lines[0].get_color()
        ax.plot(V.real, V.imag, color=color)

    x_min = V.real.min()
    x_max = V.real.max()
    y_min = V.imag.min()
    y_max = V.imag.max()

    if unitcircle:
        if x_min > -1:
            x_min = -1
        if x_max < 1:
            x_max = 1
        if y_min > -1:
            y_min = -1
        if y_max < 1:
            y_max = 1

    x_extra, y_extra = 1.0, 1.0
    x_min -= 0.5 * x_extra
    x_max += 0.5 * x_extra
    if unitcircle:
        bbox = ax.get_window_extent()
        aspect = bbox.width / bbox.height

        x_min *= aspect
        x_max *= aspect

    ax.axis('equal')
    #ax.set_xlim(x_min, x_max)

    xlabel = kwargs.pop('xlabel', 'Re')
    ylabel = kwargs.pop('ylabel', 'Im')
    title = kwargs.pop('title', None)

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    ax.grid(True)

    return ax


def plot_nichols(obj, f, norm=False, **kwargs):

    from matplotlib.pyplot import Circle, rcParams

    npoints = kwargs.pop('npoints', 400)
    # Default to True to get better plots.
    log_frequency = kwargs.pop('log_frequency', False)
    unitcircle = kwargs.pop('unitcircle', True)

    fn = None

    # FIXME, determine useful frequency range...
    if f is None:
        f = (-100, 100)
    if isinstance(f, (int, float)):
        f = (0, f)
    if isinstance(f, tuple):
        fmin, fmax = parse_range(f, 1e-1, positive=False)
        if log_frequency:
            if fmin < 0:
                frange = fmax - fmin
                f = np.geomspace(frange / 1e4, fmax, npoints // 2)
                f = np.hstack((0, f))
                fn = -np.geomspace(frange / 1e4, abs(fmin), npoints // 2)
                fn = np.hstack((0, fn))
            elif fmin == 0:
                f = np.geomspace(fmax / 1e4, fmax, npoints // 2)
                f = np.hstack((0, f))
            else:
                f = np.geomspace(fmin, fmax, npoints)
        else:
            fmin, fmax = parse_range(f, 1e-1, positive=False)
            f = np.linspace(fmin, fmax, npoints)

    V = obj.evaluate(f)
    if not V.dtype == complex:
        raise ValueError('Data not complex')

    obj = obj.doit()

    ax = make_axes(figsize=kwargs.pop('figsize', None),
                   axes=kwargs.pop('axes', None))

    lines = ax.plot(np.angle(V), 20 * np.log10(abs(V)))

    if fn is not None:
        V = obj.evaluate(fn)
        color = lines[0].get_color()
        ax.plot(np.angle(V), 20 * np.log10(abs(V)), color=color)

    xlabel = kwargs.pop('xlabel', 'Phase (radians)')
    ylabel = kwargs.pop('ylabel', 'dB')
    title = kwargs.pop('title', None)

    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    ax.grid(True)

    return ax



from lcapy import *
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_loop(self):

        a = Circuit("""
        V 1 0 {v(t)}
        R1 1 2
        R2 2 3
        R3 3 0_3
        W 0 0_3
        W 3 3_a
        R4 3_a 0_4
        W 0_3 0_4""")

        la = a.loop_analysis()

        la_eqs = la.mesh_equations()
        loops = la.loops_by_cpt_name()

        key = list(la_eqs.keys())[0]
        eq = la_eqs[key]
        loop = loops[0]

        if loop == ['R2', 'C']:
            self.assertEqual(eq.lhs, voltage(
                'R3*(I_2(t) - I_1(t)) + R4*(I_2(t) - I_1(t)) + v(t)'), 'mesh_equations()[1].lhs')
        elif loop == ['V', 'R1', 'R2', 'R3']:
            self.assertEqual(eq.lhs, voltage(
                '-R1*I_1(t) - R2*I_1(t) + R3*(-I_1(t) + I_2(t)) + v(t)'), 'mesh_equations()[1].lhs')
        elif loop == ['R3', 'R4']:
            self.assertEqual(eq.lhs, voltage(
                'R3*(I_1(t) - I_2(t)) + R4*(I_1(t) - I_2(t))'), 'mesh_equations()[1].lhs')
        else:
            raise ValueError('Unhandled loop %s' % loop)

        # Use abs since sometimes the current flows in the other direction
        # depending how the loop was chosen.

        self.assertEqual(eq.rhs, voltage(0), 'mesh_equations()[1].rhs')



"""This module provides support for the inverse z-transform.  It
calculates the unilateral inverse z-transform.

Copyright 2021-2022 Michael Hayes, UCECE

"""

from .transformer import UnilateralInverseTransformer
from .ratfun import Ratfun
from .utils import factor_const, scale_shift, pair_conjugates
from .sym import sympify, simplify, miscsymbol, AppliedUndef
from .utils import factor_const, scale_shift
from .extrafunctions import UnitImpulse, UnitStep
import sympy as sym
from sympy.simplify.fu import TR6, TR9
from warnings import warn


__all__ = ('IZT', 'inverse_ztransform')


class InverseZTransformer(UnilateralInverseTransformer):

    name = 'inverse z-transform'

    def noevaluate(self, expr, z, n):

        # Construct Cauchy integral.
        self.error('TODO')

    def check(self, expr, z, n, **kwargs):

        if expr.has(n):
            self.error('Expression depends on n')

    def key(self, expr, z, n, **kwargs):
        return (expr, z, n,
                kwargs.get('causal', False),
                kwargs.get('pairs', True),
                kwargs.get('damping', None))

    def func(self, expr, n, z, inverse=False):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        scale, shift = scale_shift(expr.args[0], n)

        # Convert v[n] to V(z), etc.
        name = expr.func.__name__
        if inverse:
            func = sym.Function(name[0].lower() + name[1:])
        else:
            func = sym.Function(name[0].upper() + name[1:])

        if not scale.is_constant():
            self.error('Cannot determine if time-expansion or decimation')

        if scale == 1:
            result = func(z)

            if shift != 0:
                result = result * z ** shift
            return result

        if scale.is_integer:
            self.error('Cannot do decimation yet')

        if not scale.is_rational:
            self.error('Cannot handle arbitrary scaling')

        if scale.p != 1:
            self.error('Cannot handle non-integer time-expansion')

        result = func(z ** scale.q)

        if shift != 0:
            result = result * z ** shift
        return result

    def ratfun(self, expr, z, n, **kwargs):

        expr = expr / z

        # Handle special case 1 / (z**m * (z - 1)) since this becomes u[n - m]
        # The default method produces u[n] - delta[n] for u[n-1].  This is correct
        # but can be simplified.
        # In general, 1 / (z**m * (z - a)) becomes a**n * u[n - m]

        if (len(expr.args) == 2 and expr.args[1].is_Pow and
            expr.args[1].args[0].is_Add and
            expr.args[1].args[0].args[0] == -1 and
                expr.args[1].args[0].args[1] == z):

            delay = None
            if expr.args[0] == z:
                delay = 1
            elif expr.args[0].is_Pow and expr.args[0].args[0] == z:
                a = expr.args[0].args[1]
                if a.is_positive:
                    warn('Dodgy z-transform 1.  Have advance of unit step.')
                elif not a.is_negative:
                    warn('Dodgy z-transform 2.  May have advance of unit step.')
                delay = -a
            elif (expr.args[0].is_Pow and expr.args[0].args[0].is_Pow and
                  expr.args[0].args[0].args[0] == z and
                  expr.args[0].args[0].args[1] == -1):
                a = expr.args[0].args[1]
                if a.is_negative:
                    warn('Dodgy z-transform 3.  Have advance of unit step.')
                elif not a.is_positive:
                    warn('Dodgy z-transform 4.  May have advance of unit step.')
                delay = a

            if delay is not None:
                return UnitStep(n - delay), sym.S.Zero

        zexpr = Ratfun(expr, z)

        Q, M, D, delay, undef = zexpr.as_QMA()

        cresult = sym.S.Zero
        uresult = sym.S.Zero

        if Q:
            Qpoly = sym.Poly(Q, z)
            C = Qpoly.all_coeffs()
            for m, c in enumerate(C):
                cresult += c * UnitImpulse(n - len(C) + m + 1)

        # There is problem with determining residues if
        # have 1/(z*(-a/z + 1)) instead of 1/(-a + z).  Hopefully,
        # simplify will fix things...
        expr = (M / D).simplify()
        # M and D may contain common factors before simplification, so redefine M and D
        M = sym.numer(expr)
        D = sym.denom(expr)
        for factor in expr.as_ordered_factors():
            if factor == sym.oo:
                return factor, factor

        zexpr = Ratfun(expr, z, **kwargs)
        poles = zexpr.poles(damping=kwargs.get('damping', None))
        poles_dict = {}
        for pole in poles:
            # Replace cos()**2-1 by sin()**2
            pole.expr = TR6(sym.expand(pole.expr))
            pole.expr = sym.simplify(pole.expr)
            # Remove abs value from sin()
            pole.expr = pole.expr.subs(sym.Abs, sym.Id)
            poles_dict[pole.expr] = pole.n

        # Juergen Weizenecker HsKa

        # Make two dictionaries in order to handle them differently and make
        # pretty expressions
        if kwargs.get('pairs', True):
            pole_pair_dict, pole_single_dict = pair_conjugates(poles_dict)
        else:
            pole_pair_dict, pole_single_dict = {}, poles_dict

        # Make n (=number of poles) different denominators to speed up
        # calculation and avoid sym.limit.  The different denominators are
        # due to shortening of poles after multiplying with (z-z1)**o
        if not (M.is_polynomial(z) and D.is_polynomial(z)):
            print("Numerator or denominator may contain 1/z terms: ", M, D)

        n_poles = len(poles)
        # Leading coefficient of denominator polynom
        a_0 = sym.LC(D, z)
        # The canceled denominator (for each (z-p)**o)
        shorten_denom = {}
        for i in range(n_poles):
            shorten_term = sym.prod(
                [(z - poles[j].expr)**(poles[j].n) for j in range(n_poles) if j != i], a_0)
            shorten_denom[poles[i].expr] = shorten_term

        # Run through single poles real or complex, order 1 or higher
        for pole in pole_single_dict:

            p = pole

            # Number of occurrences of the pole.
            o = pole_single_dict[pole]

            # X(z)/z*(z-p)**o after shortening.
            expr2 = M / shorten_denom[p]

            if o == 0:
                continue

            if o == 1:
                r = sym.simplify(sym.expand(expr2.subs(z, p)))

                if p == 0:
                    cresult += r * UnitImpulse(n)
                else:
                    uresult += r * p ** n
                continue

            # Handle repeated poles.
            all_derivatives = [expr2]
            for i in range(1, o):
                all_derivatives += [sym.diff(all_derivatives[i - 1], z)]

            bino = 1
            sum_p = 0
            for i in range(1, o + 1):
                m = o - i
                derivative = all_derivatives[m]
                # Derivative at z=p
                derivative = sym.expand(derivative.subs(z, p))
                r = sym.simplify(derivative) / sym.factorial(m)

                if p == 0:
                    cresult += r * UnitImpulse(n - i + 1)
                else:
                    sum_p += r * bino * p**(1 - i) / sym.factorial(i - 1)
                    bino *= n - i + 1

            uresult += sym.simplify(sum_p * p**n)

        # Run through complex pole pairs
        for pole in pole_pair_dict:

            p1 = pole[0]
            p2 = pole[1]

            # Number of occurrences of the pole pair
            o1 = pole_pair_dict[pole]
            # X(z)/z*(z-p)**o after shortening
            expr_1 = M / shorten_denom[p1]
            expr_2 = M / shorten_denom[p2]

            # Oscillation parameter
            lam = sym.sqrt(sym.simplify(p1 * p2))
            p1_n = sym.simplify(p1 / lam)
            # term is of form exp(j*arg())
            if len(p1_n.args) == 1 and p1_n.is_Function and p1_n.func == sym.exp:
                omega_0 = sym.im(p1_n.args[0])
            # term is of form cos() + j sin()
            elif p1_n.is_Add and sym.re(p1_n).is_Function and sym.re(p1_n).func == sym.cos:
                p1_n = p1_n.rewrite(sym.exp)
                omega_0 = sym.im(p1_n.args[0])
            # general form
            else:
                omega_0 = sym.simplify(sym.arg(p1_n))

            if o1 == 1:
                r1 = expr_1.subs(z, p1)
                r2 = expr_2.subs(z, p2)

                r1_re = sym.re(r1).simplify()
                r1_im = sym.im(r1).simplify()

                # if pole pairs is selected, r1=r2*

                # Handle real part
                uresult += 2 * TR9(r1_re) * lam ** n * sym.cos(omega_0 * n)
                uresult -= 2 * TR9(r1_im) * lam ** n * sym.sin(omega_0 * n)

            else:
                bino = 1
                sum_b = 0
                # Compute first all derivatives needed
                all_derivatives_1 = [expr_1]
                for i in range(1, o1):
                    all_derivatives_1 += [
                        sym.diff(all_derivatives_1[i - 1], z)]

                # Loop through the binomial series
                for i in range(1, o1 + 1):
                    m = o1 - i

                    # m th derivative at z=p1
                    derivative = all_derivatives_1[m]
                    r1 = derivative.subs(z, p1) / sym.factorial(m)
                    # prefactors
                    prefac = bino * lam ** (1 - i) / sym.factorial(i - 1)
                    # simplify r1
                    r1 = r1.rewrite(sym.exp).simplify()
                    # sum
                    sum_b += prefac * r1 * sym.exp(sym.I * omega_0 * (1 - i))
                    # binomial coefficient
                    bino *= n - i + 1

                # take result = lam**n * (sum_b*sum_b*exp(j*omega_0*n) + cc)
                aa = sym.simplify(sym.re(sum_b))
                bb = sym.simplify(sym.im(sum_b))
                uresult += 2 * (aa * sym.cos(omega_0 * n) -
                                bb * sym.sin(omega_0 * n)) * lam**n

        # cresult is a sum of Dirac deltas and its derivatives so is known
        # to be causal.

        return cresult, uresult

    def product(self, expr, z, n, **kwargs):

        # Handle expressions with a function of z, e.g., V(z) * Y(z), V(z)
        # / z etc.

        zsym = sympify(str(z))

        if kwargs.get('causal', False):
            # Assume that all functions are causal in the expression.
            n1 = sym.S.Zero
            n2 = n
        else:
            n1 = -sym.oo
            n2 = sym.oo

        const, expr = factor_const(expr, z)

        factors = expr.as_ordered_factors()
        if len(factors) < 2:
            self.error('Expression does not have multiple factors')

        if (len(factors) == 3 and factors[0] == z and
            isinstance(factors[2], AppliedUndef) and
            factors[1].is_Pow and factors[1].args[1] == -1 and
            factors[1].args[0].is_Add and factors[1].args[0].args[0] == -1
                and factors[1].args[0].args[1] == z):
            # Handle cumulative sum  z / (z - 1) * V(z)
            m = self.dummy_var(expr, 'm', level=0, real=True)
            result = self.func(factors[2], z, m, True)
            result = sym.Sum(result, (m, n1, n))
            return result

        # TODO, is this useful?
        if (len(factors) > 2 and not
            isinstance(factors[1], AppliedUndef) and
                isinstance(factors[2], AppliedUndef)):
            factors = [factors[0], factors[2], factors[1]] + factors[3:]

        # TODO, is this useful?
        if isinstance(factors[1], AppliedUndef):
            terms = factors[0].as_ordered_terms()
            if len(terms) >= 2:
                result = sym.S.Zero
                for term in terms:
                    result += self.product(factors[1] * term, z, n, **kwargs)
                return const * result

        cresult, uresult = self.term1(factors[0], z, n, **kwargs)
        result = cresult + uresult

        intnum = 0
        for m in range(len(factors) - 1):
            if m == 0 and isinstance(factors[1], AppliedUndef):
                # Note, as_ordered_factors puts powers of z before the functions.
                if factors[0] == z:
                    # Handle time-advance
                    # Convert z * V(z) to v[n + 1]
                    # TODO, fix for unilateral ZT
                    result = self.func(factors[1], z, n, True)
                    result = result.subs(n, n + 1)
                    continue
                elif factors[0].is_Pow and factors[0].args[0] == z and factors[0].args[1] > 0:
                    # Handle higher order advances
                    # Convert z ** k * V(z) to v[n + k]
                    result = self.func(factors[1], z, n, True)
                    result = result.subs(n, n + factors[0].args[1])
                    continue
                elif factors[0].is_Pow and factors[0].args[0] == z and factors[0].args[1] < 0:
                    # Handle time-delay  1 / z ** k * V(z)
                    result = self.func(factors[1], z, n, True)
                    result = result.subs(n, n + factors[0].args[1])
                    continue
                elif (factors[0].is_Pow and
                      factors[0].args[0].is_Add and
                      factors[0].args[1] == -1 and
                      factors[0].args[0].args[0] == 1 and
                      factors[0].args[0].args[1].is_Mul and
                      factors[0].args[0].args[1].args[0] == -1 and
                      factors[0].args[0].args[1].args[1].is_Pow and
                      factors[0].args[0].args[1].args[1].args[1] == -1 and
                      factors[0].args[0].args[1].args[1].args[0] is zsym):
                    # Handle cumulative sum  1 / (1 - 1 / z) * V(z)
                    m = self.dummy_var(expr, 'm', level=intnum, real=True)
                    result = self.func(factors[1], z, m, True)
                    intnum += 1
                    result = sym.Sum(result, (m, n1, n))
                    continue

            # Convert product to convolution
            dummy = self.dummy_var(expr, 'm', level=intnum, real=True)
            intnum += 1
            cresult, uresult = self.term1(factors[m + 1], z, n, **kwargs)
            expr2 = cresult + uresult
            kernel = result.subs(n, n - dummy) * expr2.subs(n, dummy)
            sresult = sym.Sum(kernel, (dummy, n1, n2))
            result = sresult

        return const * result

    def power(self, expr, z, n):

        # Handle expressions with a power of z.
        if (expr.is_Pow and expr.args[0] == z):
            exponent = expr.args[1]

            if exponent.is_positive:
                warn('Dodgy z-transform.  Have advance of unit impulse.')
            elif not exponent.is_negative:
                warn('Dodgy z-transform.  May have advance of unit impulse.')

            return UnitImpulse(n + exponent), sym.S.Zero

        # Handle expressions with a power of (1 / z).
        if (expr.is_Pow and expr.args[0].is_Pow and
                expr.args[0].args[0] == z and expr.args[0].args[1] == -1):

            exponent = expr.args[1]

            if exponent.is_negative:
                warn('Dodgy z-transform.  Have advance of unit impulse.')
            elif not exponent.is_positive:
                warn('Dodgy z-transform.  May have advance of unit impulse.')

            return UnitImpulse(n - exponent), sym.S.Zero

        self.error('Expression is not a power of z')

    def term1(self, expr, z, n, **kwargs):

        const, expr = factor_const(expr, z)

        if expr == 1:
            return const * UnitImpulse(n), sym.S.Zero

        if isinstance(expr, AppliedUndef):
            # Handle V(z), 3 * V(z) etc.  If causal is True it is assumed
            # that the unknown functions are causal.
            result = self.func(expr, z, n, True)
            return const * result, sym.S.Zero

        if expr.has(AppliedUndef):
            return const * self.product(expr, z, n, **kwargs), sym.S.Zero

        if expr == z:
            warn('Dodgy z-transform.  Have advance of unit impulse.')
            return const * UnitImpulse(n + 1), sym.S.Zero

        if (expr.is_Pow and
            (expr.args[0] == z or
             (expr.args[0].is_Pow and
              expr.args[0].args[0] == z and expr.args[0].args[1] == -1))):
            cresult, uresult = self.power(expr, z, n)
            return const * cresult, const * uresult

        try:
            # This is the common case.
            cresult, uresult = self.ratfun(expr, z, n)
            return const * cresult, const * uresult
        except:
            pass

        # As last resort see if can convert to convolutions...
        return sym.S.Zero, const * self.product(expr, z, n, **kwargs)

    def term(self, expr, z, n, **kwargs):

        cresult, uresult = self.term1(expr, z, n, **kwargs)

        if kwargs.get('causal', False):
            uresult = uresult * UnitStep(n)

        return cresult, uresult


inverse_ztransformer = InverseZTransformer()


def inverse_ztransform(expr, z, n, evaluate=True, **kwargs):
    """Calculate inverse z-Transform of X(s) and return x[n].

    The unilateral z-Transform transform cannot determine x[n] for n < 0
    unless given additional information:

    `dc` -- x[n] = constant
    `causal` -- x[n] = 0 for n < 0.
    `ac` -- x[n] = A cos(a * n) + B * sin(b * n)
    """

    return inverse_ztransformer.transform(expr, z, n, evaluate=evaluate,
                                          **kwargs)


def IZT(expr, z, n, evaluate=True, **kwargs):
    """Calculate inverse z-Transform of X(s) and return x[n].

    The unilateral z-Transform transform cannot determine x[n] for n < 0
    unless given additional information:

    `dc` -- x[n] = constant
    `causal` -- x[n] = 0 for n < 0.
    `ac` -- x[n] = A cos(a * n) + B * sin(b * n)
    """

    return inverse_ztransformer.transform(expr, z, n, evaluate=evaluate,
                                          **kwargs)



"""This module provides the CausalChecker, ACChecker, and DCChecker
classes as used by the is_causal, is_ac, and is_dc functions.

Copyright 2020--2021 Michael Hayes, UCECE

"""

from .sym import sympify1
from .extrafunctions import UnitImpulse, UnitStep
from .functions import Eq
import sympy as sym
from sympy import cos, pi, sin, atan2, sqrt, exp


class CausalChecker(object):

    def _has_causal_factor(self, expr):

        factors = expr.as_ordered_factors()
        for factor in factors:
            if factor == 0:
                return True
            if (not factor.is_Function
                or factor.func not in (sym.Heaviside, sym.DiracDelta,
                                       UnitImpulse, UnitStep)):
                continue

            # If have Heaviside(t), etc., then is causal
            if factor.args[0] == self.var:
                return True

            p = sym.Poly(factor.args[0], self.var)
            coeffs = p.all_coeffs()
            if len(coeffs) != 2:
                return False

            # Look for something like func(a * t + b)
            # If a positive and b negative then know func(a * t + b) = 0 for t < 0
            # TODO, generalise.
            a, b = coeffs
            if (a.is_positive and (b.is_negative or b.is_zero)):
                return True

        return False

    def _is_causal(self, expr):

        terms = expr.expand().as_ordered_terms()
        for term in terms:
            if not self._has_causal_factor(term):
                return False

        return True

    def __init__(self, expr, var):

        self.var = getattr(var, 'expr', sympify1(var))
        try:
            expr = expr.expr
        except:
            expr = sympify1(expr)

        self.is_causal = self._is_causal(expr)


class DCChecker(object):

    def _is_dc(self, expr):

        return not self.var in expr.free_symbols

    def __init__(self, expr, var):

        self.var = getattr(var, 'expr', sympify1(var))
        try:
            expr = expr.expr
        except:
            expr = sympify1(expr)

        self.is_dc = self._is_dc(expr)


class ACChecker(object):
    """This looks for real ac signals; it does not work for complex ac

signals."""

    def _find_freq_phase(self, expr):

        self.freq = 0
        self.omega = 0

        if expr.func == cos:
            self.phase = 0
        elif expr.func == sin:
            self.phase = -pi / 2
        elif expr.func == exp:
            if not expr.args[0].has(sym.I):
                return False
            self.phase = 0
            self.is_complex = True
        else:
            raise ValueError('%s not sin/cos' % expr)

        arg = expr.args[0]
        if self.is_complex:
            arg /= sym.I

        p = sym.Poly(arg, self.var)
        coeffs = p.all_coeffs()
        if len(coeffs) != 2:
            return False

        self.phase += coeffs[1]
        self.omega = coeffs[0]
        self.freq = self.omega / (2 * pi)
        return True

    def _is_sum_ac(self, terms):

        check = ACChecker(terms[0], self.var)
        if not check.is_ac:
            return False

        for term in terms[1:]:
            check2 = ACChecker(term, self.var)
            if not check2.is_ac or check.omega != check2.omega:
                return False
            if check.is_complex != check2.is_complex:
                return False
            A1, p1 = check.amp, check.phase
            A2, p2 = check2.amp, check2.phase
            x = A1 * cos(p1) + A2 * cos(p2)
            y = A1 * sin(p1) + A2 * sin(p2)
            check.phase = atan2(y, x)
            check.amp = sqrt(x**2 + y**2)

        self.freq = check.freq
        self.omega = check.omega
        self.amp = check.amp
        self.phase = check.phase
        return True

    def _is_ac(self, expr):

        # Convert exp(-j*x*t) + exp(j*x*t) into 2 * cos(x) etc.
        expr = sym.exptrigsimp(expr.rewrite(cos))

        terms = expr.as_ordered_terms()
        if len(terms) > 1:
            return self._is_sum_ac(terms)

        factors = expr.as_ordered_factors()

        self.amp = 1
        for factor in factors:
            if factor.is_Function:
                if factor.func not in (cos, sin, exp):
                    return False
                if not self._find_freq_phase(factor):
                    return False
            elif is_dc(factor, self.var):
                self.amp *= factor
            else:
                return False
        return True

    def __init__(self, expr, var):

        self.var = getattr(var, 'expr', sympify1(var))
        self.amp = 0
        self.omega = 0
        self.freq = 0
        self.phase = 0
        self.is_complex = False

        try:
            expr = expr.expr
        except:
            expr = sympify1(expr)

        self.is_ac = self._is_ac(expr)


def is_dc(expr, var):

    try:
        if expr.is_Equality:
            return is_dc(expr.expr.args[0], var) or is_dc(expr.expr.args[1], var)
    except:
        pass

    return DCChecker(expr, var).is_dc


def is_ac(expr, var):

    try:
        if expr.is_Equality:
            return is_ac(expr.expr.args[0], var) or is_ac(expr.expr.args[1], var)
    except:
        pass

    return ACChecker(expr, var).is_ac


def is_causal(expr, var):

    try:
        if expr.is_Equality:
            return is_causal(expr.expr.args[0], var) or is_causal(expr.expr.args[1], var)
    except:
        pass

    return CausalChecker(expr, var).is_causal



"""This module implements the ZDomainMatrix class for a matrix of
z-domain expressions.

Copyright 2021 Michael Hayes, UCECE

"""

from .matrix import Matrix


class ZDomainMatrix(Matrix):
    from .zexpr import ZDomainExpression
    _typewrap = ZDomainExpression

    def IZT(self, **assumptions):

        def func(expr):
            return expr.IZT(**assumptions)

        return DiscreteTimeDomainMatrix(self.applyfunc(func))

    def canonical(self):

        return self.applyfunc(self._typewrap.canonical)

    def general(self):

        return self.applyfunc(self._typewrap.general)

    def mixedfrac(self):

        return self.applyfunc(self._typewrap.mixedfrac)

    def partfrac(self):

        return self.applyfunc(self._typewrap.partfrac)

    def timeconst(self):

        return self.applyfunc(self._typewrap.timeconst)

    def ZPK(self):

        return self.applyfunc(self._typewrap.ZPK)


class ZDomainVoltageMatrix(ZDomainMatrix):
    from .zexpr import ZDomainVoltage
    _typewrap = ZDomainVoltage


class ZDomainCurrentMatrix(ZDomainMatrix):
    from .zexpr import ZDomainCurrent
    _typewrap = ZDomainCurrent


class ZDomainAdmittanceMatrix(ZDomainMatrix):
    from .zexpr import ZDomainAdmittance
    _typewrap = ZDomainAdmittance


class ZDomainImpedanceMatrix(ZDomainMatrix):
    from .zexpr import ZDomainImpedance
    _typewrap = ZDomainImpedance


from .nmatrix import DiscreteTimeDomainMatrix  # nopep8



"""This module provides the NetlistHelper class used by the
NetlistMaker and LadderMaker classes.

Copyright 2020--2022 Michael Hayes, UCECE

"""

from .componentnamer import ComponentNamer


class NetlistHelper(object):

    evalf = False

    @property
    def _node(self):

        if not hasattr(self, '_node_counter'):
            self._node_counter = 0
        ret = self._node_counter
        self._node_counter += 1
        return ret

    def _make_nodes(self, *nodes):

        return [self._node if node is None else node for node in nodes]

    def _make_name(self, cpt_type, args):
        """Make component name"""

        if not hasattr(self, '_namer'):
            self._namer = ComponentNamer()

        name = None
        # See if argument is suitable for the cpt_id
        if (len(args) > 0 and isinstance(args[0], str) and
                args[0].startswith(cpt_type) and args[0].isalnum()):
            name = args[0]

            # If have already used this name, then resort to auto name.
            if name in self.names:
                name = None

        if name is None:
            name = self._namer.name(cpt_type)

        self.names.append(name)

        return name

    def _netarg(self, arg):

        if self.evalf:
            try:
                # The arg may not be an Expr.
                arg = arg.evalf(n=self.evalf)
            except:
                pass

        try:
            # This is primarily for Superposition values.
            argstr = arg._netrepr()
        except:
            argstr = str(arg)

        # TODO: make more robust to catch expressions.
        if (('(' in argstr) or (')' in argstr) or (' ' in argstr) or
                (',' in argstr) or ('*' in argstr) or ('/' in argstr)):
            return '{%s}' % argstr
        return argstr

    def _netargs(self, net):

        return ' '.join([self._netarg(arg) for arg in net.args])



from lcapy import *
import unittest

class LcapyTester(unittest.TestCase):
    """Unit tests for lcapy"""

    def assertEqual2(self, ans1, ans2, comment):

        try:
            self.assertEqual(ans1, ans2, comment)
        except AssertionError as e:
            pprint(ans1)
            pprint(ans2)
            raise AssertionError(e)

    def test_approximate_exp(self):

        H = exp(-s)
        Ha = H.approximate_exp(order=1)
        Hb = H.approximate_exp(order=2, numer_order=1)

        self.assertEqual2(Ha, (2 - s) / (2 + s), 'approximate exp Pade order 1')
        self.assertEqual2(Hb, (6 - 2 * s) / (6 + 4 *s + s**2), 'approximate exp Pade order 1,2')

    def test_approximate_cosh(self):

        self.assertEqual2(cosh(s).approximate_hyperbolic_trig(),
                          (2 - s) / (s + 2) + (s + 2) / (2 - s),
                          'approximate cosh')

    def test_approximate_sinh(self):

        self.assertEqual2(sinh(s).approximate_hyperbolic_trig(),
                          -(2 - s) / (s + 2) + (s + 2) / (2 - s),
                          'approximate sinh')



"""This module provides discrete-time filter support.

Copyright 2021--2022 Michael Hayes, UCECE

"""

from .expr import expr, equation, ExprTuple
from .nexpr import DiscreteTimeDomainExpression
from .differenceequation import DifferenceEquation
from .discretetime import n, z, seq
from .sequence import Sequence
from .utils import isiterable
from .sym import oo
import sympy as sym


class DLTIFilter(object):

    def __init__(self, b, a):
        """Create discrete-time filter where `b` is a list or array of
        numerator coefficients and `a` is a list or array of
        denominator coefficients."""

        if not isiterable(b):
            b = (b, )
        if not isiterable(a):
            a = (a, )

        self.a = ExprTuple(a)
        self.b = ExprTuple(b)

    def __repr__(self):
        """This is called by repr(expr).  It is used, e.g., when printing
        in the debugger."""

        return '%s(%s, %s)' % (self.__class__.__name__, self.b, self.a)

    def transfer_function(self):
        """Return discrete-time impulse response (transfer function) in
        z-domain."""

        from .sym import zsym

        a = self.a.sympy
        b = self.b.sympy

        Nl = len(a)
        Nr = len(b)

        # Numerator of H(z)
        num = sym.S.Zero
        for i in range(Nr):
            num += b[i] * zsym**(-i)

        # Denominator for H(z)
        denom = a[0] * zsym**0
        for k in range(1, Nl):
            az = a[k] * zsym**(-k)
            denom += az

        # Collect with respect to positive powers of the variable z
        num = sym.collect(sym.expand(num * zsym**Nl), zsym)
        denom = sym.collect(sym.expand(denom * zsym**Nl), zsym)

        Hz = expr(sym.simplify(num / denom))
        Hz.is_causal = True
        return Hz

    def impulse_response(self):
        """Return discrete-time impulse response (transfer function) in time domain."""

        H = self.transfer_function()
        return H(n)

    def step_response(self):
        """Return discrete-time step response in time domain."""

        H = self.transfer_function()
        G = (H * z) / (z - 1)
        G.is_causal = H.is_causal

        return G(n)

    def frequency_response(self, var=None, images=oo, **assumptions):
        """Return frequency response."""

        return self.transfer_function().DTFT(var, images, **assumptions)

    def difference_equation(self, inputsym='x', outputsym='y'):
        """Return difference equation."""

        rhs = 0 * n

        for m, bn in enumerate(self.b):
            rhs += bn * expr('%s(n - %d)' % (inputsym, m))

        for m, an in enumerate(self.a[1:]):
            rhs -= an * expr('%s(n - %d)' % (outputsym, m + 1))

        lhs = self.a[0] * expr('y(n)')
        return DifferenceEquation(lhs, rhs, inputsym, outputsym)

    def zdomain_initial_response(self, ic):
        """Return z-domain response due to initial conditions.
           ic : list with initial values y[-1], y[-2], ...
        """

        from .sym import zsym

        Nl = len(self.a)
        if len(ic) != Nl:
            raise ValueError('Expecting %d initial conditions, got %d'
                             % (Nl, len(ic)))

        # Denominator for Yi(z)
        denom = self.a[0] * z**0
        num = 0 * z
        for k in range(1, Nl):
            az = self.a[k] * z**(-k)
            denom += az
            # Numerator for Yi(z)
            y0 = 0 * z
            for i in range(0, k):
                y0 += ic[i] * z**(i + 1)
            num += az * y0

        # Collect with respect to positive powers of the variable z
        num = num.sympy * zsym**Nl
        denom = denom.sympy * zsym**Nl

        num = sym.collect(sym.expand(num), zsym)
        denom = sym.collect(sym.expand(denom), zsym)

        Yzi = expr(-sym.simplify(num / denom))
        Yzi.is_causal = True

        return Yzi

    def initial_response(self, ic):
        """Return response due to initial conditions in the time domain.
           ic : list with initial values y[-1], y[-2], ...
        """

        Yzi = self.zdomain_initial_response(ic)
        return Yzi(n)

    def response(self, x, ic=None, ni=None):
        """Calculate response of filter to input `x` given a list of initial conditions
        `ic` for time indexes specified by `ni`.  If `ni` is a tuple,
        this specifies the first and last (inclusive) time index.

        The initial conditions are valid prior to the time indices given by the ni
        `x` can be an expression, a sequence, or a list/array of values.
        """

        from numpy import arange, ndarray

        if (ic is None and ni is None and isinstance(x, ndarray)
                and self.a.symbols == {} and self.b.symbols == {}):

            from scipy.signal import lfilter

            # Calculate numerical response
            b = self.b.fval
            a = self.a.fval
            return lfilter(b, a, x)

        if ic is None:
            ic = [0] * (len(self.a) - 1)

        if not isiterable(x) and not isinstance(x, DiscreteTimeDomainExpression):
            x = (x, )

        if not isiterable(ic):
            ic = (ic, )

        if isinstance(x, (tuple, list, ndarray)):
            x = seq(x)
        elif not isinstance(x, (Sequence, DiscreteTimeDomainExpression)):
            raise ValueError(
                'The input x must be a scalar, tuple, sequence, nexpr, list, or array')

        NO = len(ic)

        if NO != len(self.a) - 1:
            raise ValueError(
                "Expected %d initial conditions, got %d" % (len(self.a) - 1, NO))

        if ni is None:
            ni = (0, 10)

        if isinstance(ni, tuple):
            ni = arange(ni[0], ni[1] + 1)

        Nn = len(ni)

        # Order right hand side
        Nr = len(self.b)

        y_tot = list(ic[-1::-1]) + Nn * [0]

        a_r = self.a[-1:-1-NO:-1]
        for i, nval in enumerate(ni):
            # Get previous y vals (sliding window)
            pre_y = y_tot[i:i + NO]

            # Calculate rhs of new value
            if isinstance(x, Sequence):
                rhs = sum(self.b[l] * x[nval - l] for l in range(Nr))
            else:
                rhs = sum(self.b[l] * x(nval - l) for l in range(Nr))

            # Add lhs
            y_tot[i + NO] = -1 / self.a[0] * \
                sum(csi * ysi for csi, ysi in zip(a_r, pre_y)) + \
                rhs / self.a[0]

        # Solution, without initial values
        ret_seq = seq(y_tot[NO:], ni)

        return ret_seq

    def subs(self, *args, **kwargs):

        a = self.a.subs(*args, **kwargs)
        b = self.b.subs(*args, **kwargs)

        return self.__class__(b, a)

    def pdb(self):
        """Enter the python debugger."""

        import pdb
        pdb.set_trace()
        return self

    @property
    def is_stable(self):
        """Return True if all the poles of the filters transfer function
        are within the unit circle."""

        return self.transfer_function().is_stable



"""This module provides the PhasorDomain classes to represent phasors
 for AC analysis.

A phasor represents the amplitude and phase for a single sinusoid.  By
default the angular frequency is omega_0 but it can be any number or
symbol.

Phasors straddle the time and frequency domains and hence the
confusing phasor classes.

A phasor is described by an amplitude and phase.  There is also an
implicit frequency.  The amplitude and phase are usually functions of
frequency.  A phasor is useful to describe an AC voltage or current
signal.

A ratio of two phasors (of the same frequency) is no longer a
time-domain signal but a frequency domain quantity.  The frequency
dependence is explicit.  A phasor ratio is useful to describe an
immittance or transfer function.  These are frequency domain concepts.
A phasor ratio can be inferred from the Laplace domain by substituting
jomega (or jw) for s, where omega is the angular frequency of the phasor.

Lcapy considers V(jw) or I(jw) a generic phasor but V(3j) or I(3j) a
specific phasor.  Similarly, X(jw) is a generic phasor ratio and X(3j)
is a specific phasor ratio, where X denotes an immittance or transfer
function.

Copyright 2014--2021 Michael Hayes, UCECE

"""

from __future__ import division
from .expressionclasses import expressionclasses
from .acdc import ACChecker
from .domains import PhasorTimeDomain, PhasorFrequencyDomain
from .sym import j, omegasym
from .expr import expr
from .functions import sin, cos, exp, sqrt
from .expr import Expr
from .omegaexpr import omega
from sympy import Expr as symExpr
from warnings import warn


__all__ = ('phasor', )

# The phasor domain is different from the Fourier and Laplace domain
# since there is an implicit angular frequency.  This is only needed
# for voltage and current (vorrent) expressions.

# The phasor domain immittance can be found from the Laplace domain
# impedance/admittance by substituting s with j * omega.  In many
# cases the result is the same as the Fourier domain immittance but
# not always.  The most important case is the impedance of a
# capacitor.  In the Laplace domain it is 1/(s C); in the phasor
# domain it is 1/(j omega C); in the Fourier domain it is 1/(2 j omega
# C) + delta(omega) / (2 C).


class PhasorExpression(Expr):

    @property
    def omega(self):
        """Return angular frequency."""

        return self.assumptions.get('omega', None)

    @property
    def var(self):
        """Return underlying variable as sympy expression."""

        # TODO, think this through...
        var = self.omega
        if hasattr(var, 'expr'):
            raise ValueError('FIXME, omega is Lcapy symbol')

        # Handle things like 2 * pi * f
        if isinstance(var, symExpr) and var.is_Mul:
            return None

        return var

    def phasor(self, **assumptions):
        """Convert to phasor representation."""

        ass = self.assumptions.copy()
        assumptions = ass.merge(**assumptions)

        return self.__class__(self, **assumptions)

    def rms(self):
        """Return root mean square."""

        return abs(self) * sqrt(2) / 2

    def plot(self, wvector=None, **kwargs):
        """Plot polar diagram for a time-domain phasor or frequency response
        for a frequency-domain phasor.  For the latter, wvector
        specifies the angular frequencies.  If it is a tuple, it sets
        the angular frequency limits."""

        from .plot import plot_phasor, plot_angular_frequency

        if self.is_phasor_time_domain:
            return plot_phasor(self, **kwargs)

        if self.omega.is_constant():
            raise ValueError('Cannot plot at single frequency')

        return plot_angular_frequency(self, wvector, **kwargs)

    def bode_plot(self, fvector=None, unwrap=True, **kwargs):
        """Plot frequency response for a frequency-domain phasor as a Bode
        plot (but without the straight line approximations).  fvector
        specifies the frequencies.  If it is a tuple (m1, m2), it sets the
        frequency limits as (10**m1, 10**m2).

        `unwrap` controls phase unwrapping (default True).
        """

        from .plot import plot_bode

        if not self.is_phasor_frequency_domain:
            raise ValueError('Not frequency domain phasor: use plot()')

        result = self.fourier()
        return plot_bode(result, fvector, unwrap=unwrap, **kwargs)

    def _mul_compatible(self, x):

        if not hasattr(x, 'omega'):
            return True

        if self.omega == x.omega:
            return True

        raise ValueError('Incompatible phasor angular frequencies %s and %s' %
                         (self.omega, x.omega))

    def subs(self, *args, safe=False, **kwargs):
        """Substitute variables in expression, see sympy.subs for usage."""

        from .sym import fsym, pi

        # Handle omega -> 2 * pi * f conversions to Fourier domain.
        if len(args) == 1 and args[0] == 2 * pi * fsym:
            from .fexpr import FourierDomainExpression

            if self.var != omegasym:
                raise ValueError('Expecting omega for self.var')

            # Could check for expressions that are known to fail,
            # such as 1 / (j * omega).
            if not safe:
                warn("""
Converting to Fourier domain via phasor domain may not give correct answer.   It is safer to convert to time domain then to Fourier domain.""")

            return FourierDomainExpression(self.sympy.subs(omegasym, args[0].sympy)).as_quantity(self.quantity)

        return super(PhasorExpression, self).subs(*args, **kwargs)


class PhasorTimeDomainExpression(PhasorTimeDomain, PhasorExpression):
    """This is a phasor domain base class for voltages and currents."""

    is_phasor_domain = True
    is_phasor_time_domain = True

    def __init__(self, val, **assumptions):

        if isinstance(val, PhasorExpression):
            assumptions['omega'] = val.omega
        elif 'omega' not in assumptions or assumptions['omega'] is None:
            assumptions['omega'] = omegasym

        if hasattr(assumptions['omega'], 'expr'):
            assumptions['omega'] = assumptions['omega'].expr

        assumptions['ac'] = True
        super(PhasorExpression, self).__init__(val, **assumptions)

    def _class_by_quantity(self, quantity, domain=None):

        if quantity == 'undefined':
            return PhasorTimeDomainExpression

        return super(PhasorTimeDomainExpression, self)._class_by_quantity(quantity, domain)

    def as_expr(self):
        return PhasorTimeDomainExpression(self)

    @classmethod
    def from_time(cls, expr, omega=None, **assumptions):

        from .symbols import t

        if expr.is_admittance or expr.is_impedance or expr.is_transfer:
            warn('Should convert %s expression to Laplace-domain first.' %
                 expr.quantity)

        assumptions['ac'] = True

        if expr.is_transform_domain:
            warn('Converting %s-domain to time-domain first.' % expr.domain)
            expr = expr.time()

        check = ACChecker(expr, t)
        if not check.is_ac:
            raise ValueError(
                'Do not know how to convert %s to phasor.  Expecting an AC signal.' % expr)

        if omega is not None and check.omega != omega:
            raise ValueError('Expecting omega=%s, found omega=%s.' %
                             (omega, check.omega))

        if check.omega == 0:
            warn('DC phasor.')

        result = check.amp * exp(j * check.phase)
        assumptions['omega'] = check.omega
        assumptions['complex_signal'] = check.is_complex

        return cls.change(expr, result, domain='phasor', **assumptions)

    def time(self, **assumptions):
        """Convert to time domain representation."""
        from .symbols import t

        omega1 = self.omega
        if isinstance(omega1, Expr):
            # TODO: Fix inconsistency.  Sometimes omega is a symbol.
            omega1 = omega1.expr

        if self.is_complex_signal:
            result = self.expr * exp(j * omega1 * t)
        else:
            result = self.real.expr * \
                cos(omega1 * t) - self.imag.expr * sin(omega1 * t)

        return TimeDomainExpression(result).as_quantity(self.quantity)

    def laplace(self, **assumptions):
        """Convert to Laplace domain representation."""

        return self.time().laplace()

    def angular_fourier(self, **assumptions):
        """Angular Fourier transform."""

        if self.has(omega):
            warn('Expression contains omega, should substitute with a different symbol.')

        return self.time().angular_fourier()

    def fourier(self, **assumptions):
        """Fourier transform."""

        return self.time().fourier()


class PhasorFrequencyDomainExpression(PhasorFrequencyDomain, PhasorExpression):
    """This represents the ratio of two-phasors; for example
    an impedance, an admittance, or a transfer function."""

    is_phasor_frequency_domain = True
    is_transform_domain = True

    def __init__(self, val, **assumptions):

        if isinstance(val, PhasorExpression):
            assumptions['omega'] = val.omega
        elif 'omega' not in assumptions or assumptions['omega'] is None:
            assumptions['omega'] = omegasym

        if isinstance(val, Expr):
            ass = val.assumptions.copy()
            ass = ass.merge(**assumptions)
        else:
            ass = assumptions

        super(PhasorExpression, self).__init__(val, **ass)

    def _class_by_quantity(self, quantity, domain=None):

        if quantity == 'undefined':
            return PhasorFrequencyDomainExpression

        return super(PhasorFrequencyDomainExpression, self)._class_by_quantity(quantity, domain)

    @classmethod
    def from_laplace(cls, expr, **assumptions):

        from .sym import ssym

        ass = expr.assumptions.copy()
        ass = ass.merge(**assumptions)
        omega = ass.pop('omega', None)

        if omega is None:
            omega = omegasym
        elif hasattr(omega, 'expr'):
            omega = omega.expr

        if expr.is_voltage or expr.is_current:
            warn('Should convert %s expression to time-domain first.' %
                 expr.quantity)

        # Substitute jw for s
        result = expr.laplace(**ass)
        result2 = result.expr.replace(ssym, j * omegasym)

        quantity = expr.quantity
        if quantity == 'undefined':
            cls = PhasorFrequencyDomainExpression
        else:
            cls = expr._class_by_domain('phasor')

        ret = cls(result2, omega=omega, **ass)
        return ret

    def time(self, **assumptions):
        """Convert to time domain representation."""

        return self.laplace().time()

    def laplace(self, **assumptions):
        """Convert to Laplace domain representation."""

        from .sym import ssym
        from .sexpr import LaplaceDomainExpression

        ass = self.assumptions.copy()
        assumptions = ass.merge(**assumptions)

        omega = self.omega
        result = self.expr.replace(omega, ssym / j)
        return LaplaceDomainExpression(result, **assumptions).as_quantity(self.quantity)

    def angular_fourier(self, **assumptions):
        """Angular Fourier transform."""

        if self.var is omega:
            # TODO, add warning.
            pass

        return self.time().angular_fourier()

    def fourier(self, **assumptions):
        """Fourier transform."""

        return self.time().fourier()

    def as_expr(self):
        return PhasorFrequencyDomainExpression(self)


def phasor(arg, omega=None, **assumptions):
    """Create phasor.

    If arg has the form A * cos(w * t + phi) the phasor
    A * exp(j * phi) of angular frequency w is returned.

    If arg has the form A * sin(w * t + phi) the phasor
    A * exp(j * (phi - pi / 2)) of angular frequency w is returned.

    If arg is a constant C, the phasor C is created with omega as the
    angular frequency (this defaults to 'omega' if not specified).

    """

    arg = expr(arg)

    if arg.is_time_domain:
        # Expecting AC signal.
        return PhasorTimeDomainExpression.from_time(arg, omega=omega, **assumptions)
    elif arg.is_unchanging:
        # Expecting phasor (complex amplitude)
        return PhasorTimeDomainExpression(arg, omega=omega, **assumptions)
    else:
        # Is this sensible?  It is probably better to have
        # a phasorratio function.  TODO add deprecation warning.
        return PhasorFrequencyDomainExpression(arg, omega=omega, **assumptions)


expressionclasses.register('phasor', PhasorTimeDomainExpression,
                           PhasorFrequencyDomainExpression,
                           ('voltage', 'current',
                            'voltagesquared', 'currentsquared', 'undefined'))

from .texpr import TimeDomainExpression  # nopep8
from .expr import Expr  # nopep8



from matplotlib.pyplot import savefig
from lcapy import j, f, degrees, pi

f1 = 10
f2 = 1e6
A0 = 1e6

A = A0 * (1 / (1 + j * f / f1)) * (1 / (1 + j * f / f2))

ax = A.bode_plot((1, 10e6), plot_type='dB-degrees')
ax[0].set_ylim(-30, 130)
ax[1].set_ylim(-240, 0)

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



from numpy import zeros, hstack, dot, argsort, min, max, vstack
from scipy.linalg import inv, lu
from .cnodes import Cnodes
from .schemplacerbase import SchemPlacerBase
from warnings import warn

# This is an experimental component placement algorithm for schematics
# based on solving a system of linear equations.  The unknowns are the
# node positions and the stretch requried for stretchy components.
#
# The algorithm identifies the free stretch variables and assigns them
# to be zero.  However, this depends on the constraint ordering and
# so sometimes a negative stretch is generated due to a poor choice
# of basic variable.  This could be avoided with a constrained
# optimisation where the node positions and component stretches all
# need to be non-negative.


class Constraint(object):

    def __init__(self, size, stretch):
        self.size = size
        self.stretch = stretch

    def __repr__(self):
        return '%.2f%s' % (self.size, '*' if self.stretch else '')


class Constraints(dict):

    def __repr__(self):

        s = ''
        for key, constraint in self.items():
            s += '%s -> %s: %s\n' % (key[0], key[1], repr(constraint))
        return s


class Lineq(object):

    def __init__(self, direction, nodes, debug=0):

        self.direction = direction
        self.cnodes = Cnodes(nodes)
        self.debug = debug
        self.constraints = Constraints()

    def link(self, n1, n2):
        """Make nodes n1 and n2 share common node"""

        self.cnodes.link(n1, n2)

    def add(self, elt, n1, n2, size, stretch):

        if size == 0:
            return

        if size < 0:
            n1, n2 = n2, n1
            size = -size

        if n1 in self.cnodes:
            n1 = self.cnodes[n1]
        if n2 in self.cnodes:
            n2 = self.cnodes[n2]

        key = n1, n2
        key2 = n2, n1
        if key not in self.constraints and key2 not in self.constraints:
            self.constraints[key] = Constraint(size, stretch)
            return

        if key2 in self.constraints:
            size = -size
            key = key2
        constraint = Constraint(size, stretch)

        constraint2 = self.constraints[key]
        if not constraint2.stretch:
            if (not constraint.stretch and constraint2.size != constraint.size):
                raise ValueError('Incompatible fixed constraint of size %s and %s' % (
                    constraint.size, constraint2.size))
            self.constraints[key] = constraint
            return

        if abs(constraint.size) > abs(constraint2.size):
            self.constraints[key] = constraint

    def solve(self, stage=None):

        cnode_map = {}
        cnode_imap = []
        num = 0
        for node, cnode in self.cnodes.items():
            if cnode not in cnode_map:
                cnode_map[cnode] = num
                cnode_imap.append(cnode)
                num += 1

        Nstretches = 0
        for key, constraint in self.constraints.items():
            if constraint.stretch:
                Nstretches += 1

        Nnodes = len(cnode_map)

        if Nnodes == 1:
            pos = {}
            for node, cnode in self.cnodes.items():
                m = cnode_map[self.cnodes[node]]
                pos[node] = 0
            return pos, 0

        Nunknowns = Nnodes - 1 + Nstretches
        Nconstraints = len(self.constraints)

        A = zeros((Nconstraints, Nunknowns))
        b = zeros((Nconstraints, 1))

        m = 0
        s = 0

        for key, constraint in self.constraints.items():
            m1 = cnode_map[key[0]]
            m2 = cnode_map[key[1]]
            size = constraint.size

            if m1 != 0:
                A[m, m1 - 1] = -1
            if m2 != 0:
                A[m, m2 - 1] = 1
            if constraint.stretch:
                A[m, s + Nnodes - 1] = -1
                s += 1

            b[m] = size
            m += 1

        # Augmented matrix.
        W = hstack((A, b))

        # Extract upper triangular form (this is similar to
        # reduce row-echelon form (RREF) but not always).
        PL, U = lu(W, permute_l=True)

        # Non-zero for bound (basic) variables.
        bound = zeros(W.shape[-1], dtype=int)

        for r in range(Nnodes - 1):
            if U[r, r] == 0:
                raise ValueError(
                    'No basic variable for node %s' % cnode_imap[r])
            bound[r] = 1

        for r in range(U.shape[0]):
            if U[r, r] != 0:
                bound[r] = 1
                # For prettiness
                if U[r, r] < 0:
                    U[r, r:] = -U[r, r:]

        Nbasic = bound.sum()
        Ur = U[0:Nbasic, bound != 0]

        br = U[0:Nbasic, -1]
        # A slow one-liner!  Just need to back-substitute.
        xx = dot(inv(Ur), br)
        x = xx[0: Nnodes - 1]

        # A negative stretch can be generated by inconsistent constraints
        # or more likely by a poor choice of a stretch as a basic variable.
        # This depends on the constraint ordering since the first stretch
        # unknowns will be selected.
        for m in range(Nnodes - 1, xx.shape[0]):
            if xx[m] < 0:
                warn('Negative stretch %s' % xx[m])

        minx = min(x)
        width = max(x) - minx

        pos = {}
        for node, cnode in self.cnodes.items():
            m = cnode_map[self.cnodes[node]]
            if m == 0:
                pos[node] = 0 - minx
            else:
                pos[node] = x[m - 1] - minx

        self.W = W
        self.b = b
        self.U = U

        if ((self.direction == 'horizontal' and self.debug & 4)
                or (self.direction == 'vertical' and self.debug & 8)):

            if self.debug & 16:
                from .matrix import Matrix
                W = Matrix(W)
                print(W.latex())

            if self.debug & 32:
                from .matrix import Matrix
                U = Matrix(U)
                print(U.latex())

            if self.debug & 64:
                for key, cnode in self.cnodes.items():
                    print(key, cnode)

            if self.debug & 128:
                for key, constraint in self.constraints.items():
                    print(key, constraint)

            if self.debug & 256:
                import pdb
                pdb.set_trace()

        return pos, width


class SchemLineqPlacer(SchemPlacerBase):

    def __init__(self, elements, nodes, debug=0):

        self.elements = elements
        self.nodes = nodes
        self.debug = debug

        self.xgraph = Lineq('horizontal', nodes, debug)
        self.ygraph = Lineq('vertical', nodes, debug)

        warn('This is experimental')



"""This module provides the SequenceExpression class to provide
common methods for the discrete-time and discrete-frequency
expressions.

Copyright 2020--2022 Michael Hayes, UCECE

"""

from .dexpr import DiscreteExpression
from .sequence import Sequence
from .functions import function_mapping
from .extrafunctions import UnitStep, UnitImpulse, dtrect, dtsign
from sympy import Heaviside
from warnings import warn


class SequenceExpression(DiscreteExpression):
    """Superclass of discrete-time and discrete-frequency expressions."""

    def __init__(self, val, **assumptions):

        super(SequenceExpression, self).__init__(val, **assumptions)

        def remap_continuous_discrete(expr):

            def query(expr):
                return expr.is_Function and expr.func in function_mapping

            def value(expr):
                # Note, Heaviside behaviour changed in SymPy-1.9
                # The second arg now defaults to 1/2, so we need
                # to remove it.
                if expr.func == Heaviside:
                    return UnitStep(expr.args[0])
                return function_mapping[expr.func](*expr.args)

            # Use discrete-time function variants, see also functions.py
            return expr.replace(query, value)

        self.expr = remap_continuous_discrete(self.expr)

    def first_index(self, ni=None):

        if ni is None:
            ni = (-10, 10)
        if isinstance(ni, tuple):
            ni = range(*ni)

        # Desire SymPy equivalent to argmax
        n2 = ni[-1]
        for n in reversed(ni):
            if self(n) != 0:
                n2 = n
        return n2

    def last_index(self, ni=None):

        if ni is None:
            ni = (-10, 10)
        if isinstance(ni, tuple):
            ni = range(*ni)

        # Desire SymPy equivalent to argmin
        n1 = ni[0]
        for n in ni:
            if self(n) != 0:
                n1 = n
        return n1

    def seq(self, ni=None, evaluate=False):
        """Create a Sequence.

        >>> a = x.seq()

        The sequence indices are specified with the optional `ni` argument.
        For example:

        >>> a = x.seq(ni=(-1, 0, 1, 2))

        If the `ni` argument is not specified, the sequence indices
        are enumerated from 0.

        The sequence indices can be found using the `n` attribute.
        This returns a list.

        >>> a = x.seq().n
        [-1, 0, 1, 2]
        """

        from numpy import arange

        start_trunc = False
        end_trunc = False

        if ni is None:
            n1 = self.first_index(ni)
            n2 = self.last_index(ni)
            ni = arange(n1, n2 + 1)
            # Should search to handle cases such as 1, 1, 0, 1 but
            # when to stop?
            if self(n1 - 1) != 0:
                warn('Sequence truncated at n1=%d' % n1)
                start_trunc = True
            if self(n2 + 1) != 0:
                warn('Sequence truncated at n2=%d' % n2)
                end_trunc = True

        elif isinstance(ni, tuple):
            ni = arange(ni[0], ni[-1] + 1)

        v = self(ni)

        return self.seqcls(v, ni, evaluate=evaluate,
                           start_trunc=start_trunc, end_trunc=end_trunc)

    @property
    def is_stable(self):
        """Return True if all the poles of the signal's z-transform
        are within the unit circle."""

        poles = self.poles(aslist=True)
        for pole in poles:
            if abs(pole) >= 1:
                return False
        return True



"""
This module performs state-space analysis.

Copyright 2021-2022 Michael Hayes, UCECE

"""

from .mnacpts import I, V, C, L
from .expr import expr
from .matrix import Matrix
from .tmatrix import TimeDomainMatrix
from .voltage import voltage
from .current import current
from .statespace import StateSpace
from .circuitgraph import CircuitGraph
from .sym import sympify
import sympy as sym
from warnings import warn

__all__ = ('StateSpaceMaker', )

# TODO
# 1. Use a better Matrix class that preserves the class of each
# element, where possible.
# 2. Handle namespaces

# Independent sources
# V1 1 0
# V1 1 0 {10 * u(t)}
# V1 1 0 ac 10
#
# In each case, we use v1(t) as the independent source when determining
# the A, B, C, D matrices.  We can then substitute the known value at
# the end.

# There is a generalised (but less common) state-space representation:
#
# dx/dt = A x + B u + C du/dt
# y = D x + E u + F du/dt
#
# This is required when have a current source in series with an inductor;
# here the inductor current is not a state variable since it is equivalent
# to an input.  Moreover, the output equation for the inductor voltage
# requires the derivative of the input current.


def _hack_vars(exprs):
    """Substitute i_Canon1(t) with i_C(t) etc. provided
    there is no i_Canon2(t)."""

    for m, expr1 in enumerate(exprs):
        for c in ('i_V', 'i_C', 'i_L', 'v_C'):
            sym1 = sympify(c + 'anon1(t)')
            sym2 = sympify(c + 'anon2(t)')
            if expr1.has(sym1) and not expr1.has(sym2):
                expr1 = expr1.subs(sym1, sympify(c + '(t)'))
                exprs[m] = expr1


class StateSpaceMaker(object):
    """This converts a circuit to state-space representation.

    It can take a long time for a symbolic circuit with many reactive
    components.

    The currents through inductors and the voltage across capacitors
    are chosen as the state variables.  However, this fails when a
    current source is in series with an inductor.  In this case, the
    inductor current is not a state variable since it is equivalent to
    an input.  Moreover, the output equation for the inductor voltage
    requires the derivative of the input current.

    This does not (yet) look for degenerate circuits.  These are
    circuits with a loop consisting only of voltage sources and/or
    capacitors, or a cut set consisting only of current sources and/or
    inductors.  One hack is to call simplify() first to remove series
    inductors and parallel capacitors.

    """

    @classmethod
    def from_circuit(cls, cct, node_voltages=None, branch_currents=None):
        """`node_voltages` is a list of node names to use as voltage outputs.
        If `None` use all the unique node names.  Use `()`
        if want no branch currents.

        `branch_currents` is a list of component names to use as
        current outputs.  If `None` use all the components.  Use `()`
        if want no branch currents."""

        if node_voltages is None:
            node_voltages = cct.node_list

        if branch_currents is None:
            branch_currents = cct.branch_list

        if node_voltages == [] and branch_currents == []:
            raise ValueError('State-space: no outputs')

        inductors = []
        capacitors = []
        independent_current_sources = []
        independent_voltage_sources = []

        # Determine state variables (current through inductors and
        # voltage across acapacitors) and replace inductors with
        # current sources and capacitors with voltage sources.
        sscct = cct._new()
        cpt_map = {}

        for key, elt in cct.elements.items():
            ssnet = elt._ss_model()
            sselt = sscct._add(ssnet)
            name = elt.name
            cpt_map[name] = sselt.name

            if elt.is_inductor:
                if sselt.name in cct.elements:
                    raise ValueError(
                        'Name conflict %s, either rename the component or improve the code!' % sselt.name)

                inductors.append(elt)
            elif elt.is_capacitor:
                if sselt.name in cct.elements:
                    raise ValueError(
                        'Name conflict %s, either rename the component or improve the code!' % sselt.name)
                capacitors.append(elt)
            elif elt.is_independent_current_source:
                independent_current_sources.append(elt)
            elif elt.is_independent_voltage_source:
                independent_voltage_sources.append(elt)

        independent_sources = independent_voltage_sources + independent_current_sources

        # Build circuit graph and check if have inductor in series with
        # current source.
        if independent_current_sources != [] and inductors != []:
            cg = CircuitGraph(cct)
            for elt in independent_current_sources:
                for name in cg.in_series(elt.name):
                    if cct[name].is_inductor:
                        raise ValueError(
                            'Cannot create state-space model since have inductor %s in series with independent current source %s' % (name, elt.name))

        cct = cct
        sscct = sscct
        # sscct can be analysed in the time domain since it has no
        # reactive components.  However, for large circuits
        # this can take a long time due to inversion of the MNA matrix.

        try:
            # Analyse node voltages and branch currents.
            sscct[0].V
        except ValueError as e:
            reasons = []
            if len(inductors) > 0:
                reasons.append(
                    'Check for cut set consisting only of current sources and/or inductors.')
            if len(capacitors) > 0:
                reasons.append(
                    'Check for a loop consisting of voltage sources and/or capacitors.')

            raise ValueError("State-space analysis failed.\n%s\n     %s" % (
                e, '\n    '.join(reasons)))

        dotx_exprs = []
        statevars = []
        statenames = []
        initialvalues = []
        for elt in inductors + capacitors:
            name = cpt_map[elt.name]

            if isinstance(elt, L):
                # Inductors  v = L di/dt  so need v across the L
                expr = sscct[name].v / elt.cpt.L
                var = -sscct[name].isc
                x0 = elt.cpt.i0
            else:
                # Capacitors  i = C dv/dt  so need i through the C
                # The current is negated since it is from a source V_Cx
                expr = -sscct[name].i / elt.cpt.C
                var = sscct[name].voc
                x0 = elt.cpt.v0

            dotx_exprs.append(expr)
            statevars.append(var)
            statenames.append(name)
            initialvalues.append(x0)

        statesyms = sympify(statenames)

        # Determine independent sources.
        sources = []
        sourcevars = []
        sourcenames = []
        for elt in independent_sources:
            name = cpt_map[elt.name]

            if isinstance(elt, V):
                expr = elt.cpt.voc
                var = sscct[name].voc
            else:
                expr = elt.cpt.isc
                var = sscct[name].isc

            sources.append(expr)
            sourcevars.append(var)
            sourcenames.append(name)

        sourcesyms = sympify(sourcenames)

        # linear_eq_to_matrix expects only Symbols and not AppliedUndefs,
        # so substitute.
        subsdict = {}
        for var, sym1 in zip(statevars, statesyms):
            subsdict[var.expr] = sym1
        for var, sym1 in zip(sourcevars, sourcesyms):
            subsdict[var.expr] = sym1

        for m, expr in enumerate(dotx_exprs):
            dotx_exprs[m] = expr.subs(subsdict).expr.expand()

        if dotx_exprs == []:
            warn('State-space: no state variables found')
        if sourcesyms == []:
            warn('State-space: no independent sources found')

        if dotx_exprs != []:
            A, b = sym.linear_eq_to_matrix(dotx_exprs, *statesyms)
        else:
            A = sym.zeros(len(dotx_exprs), len(dotx_exprs))

        if sourcesyms != [] and dotx_exprs != []:
            B, b = sym.linear_eq_to_matrix(dotx_exprs, *sourcesyms)
        else:
            B = sym.zeros(len(dotx_exprs), len(sources))

        # Determine output variables.
        yexprs = []
        y = []

        for node in node_voltages:
            if node == '0':
                continue
            yexprs.append(sscct[node].v.subs(subsdict).expand().expr)
            # Note, this can introduce a name conflict
            y.append(voltage('v_%s(t)' % node))

        for name in branch_currents:
            # Perhaps ignore L since the current through it is a
            # state variable?
            name2 = cpt_map[name]
            yexprs.append(sscct[name2].i.subs(subsdict).expand().expr)
            y.append(current('i_%s(t)' % name))

        if statesyms != [] and yexprs != []:
            C, b = sym.linear_eq_to_matrix(yexprs, *statesyms)
        else:
            C = sym.zeros(len(yexprs), len(statesyms))

        if sourcesyms != [] and yexprs != []:
            D, b = sym.linear_eq_to_matrix(yexprs, *sourcesyms)
        else:
            D = sym.zeros(len(yexprs), len(sourcesyms))

        # Rewrite vCanon1(t) as vC(t) etc if appropriate.
        _hack_vars(statevars)
        _hack_vars(sources)

        # Note, Matrix strips the class from each element...
        x = TimeDomainMatrix(statevars)

        x0 = Matrix(initialvalues)

        u = TimeDomainMatrix(sources)

        # Perhaps could use v_R1(t) etc. as the output voltages?
        y = TimeDomainMatrix(y)

        A = Matrix(A)
        B = Matrix(B)
        C = Matrix(C)
        D = Matrix(D)

        return StateSpace(A, B, C, D, u, y, x, x0)



"""This module implements the DiscreteTimeDomainMatrix class for a matrix of
time-domain expressions.

Copyright 2021 Michael Hayes, UCECE

"""

from .matrix import Matrix


class DiscreteTimeDomainMatrix(Matrix):
    from .nexpr import DiscreteTimeDomainExpression
    _typewrap = DiscreteTimeDomainExpression

    def ZT(self):

        def func(expr):
            return expr.ZT()

        return ZDomainMatrix(self.applyfunc(func))


from .zmatrix import ZDomainMatrix  # nopep8
